from django.forms import widgets

from django.utils.encoding import force_unicode
from django.forms.util import flatatt
from django.utils.safestring import mark_safe
from django.utils.html import escape
from django.utils import simplejson as json

from dockit.forms.fields import HiddenJSONField

from urllib import urlencode

class DotPathWidget(widgets.Input):
    input_type = 'submit'
    
    def __init__(self, dotpath=None):
        self.dotpath = dotpath
        super(DotPathWidget, self).__init__()
    
    def render_button(self, dotpath, edit=False):
        if edit:
            label = 'edit'
        else:
            label = 'add'
        data = {'next_dotpath':dotpath}
        name = '[fragment]%s' % urlencode(data)
        submit_attrs = self.build_attrs({}, type=self.input_type, name=name, value=label)
        return mark_safe(u'<input%s />' % flatatt(submit_attrs))
    
    def prep_value(self, value):
        if isinstance(value, list):
            new_values = list()
            for item in value:
                if hasattr(item, 'to_primitive'):
                    item = item.to_primitive(item)
                new_values.append(item)
            value = new_values
        if hasattr(value, 'to_primitive'):
            value = value.to_primitive(value)
        if value is None:
            return ''
        elif not isinstance(value, basestring): 
            return json.dumps(value)
        return value
    
    def get_label(self, dothpath, value=None):
        if value:
            return escape(force_unicode(value))
        return ''
    
    def render(self, name, value, attrs=None):
        path_parts = list()
        if self.dotpath:
            path_parts.append(self.dotpath)
        path_parts.append(name) #TODO consider this will break if there is a prefix!
        dotpath = '.'.join(path_parts)
        json_value = self.prep_value(value)
        
        data_attrs = self.build_attrs(attrs, type='hidden', name=name, value=json_value)
        data_html = mark_safe(u'<input%s />' % flatatt(data_attrs))
        
        if isinstance(value, list):
            rows = list()
            index = -1
            for index, item in enumerate(value):
                item_dotpath = '%s.%s' % (dotpath, index)
                butn_html = self.render_button(item_dotpath, edit=True)
                rows.append('<td>%s</td><td>%s</td>' % (self.get_label(item_dotpath, item), butn_html))
            item_dotpath = '%s.%s' % (dotpath, index+1)
            butn_html = self.render_button(item_dotpath)
            rows.append('<td>%s</td><td>%s</td>' % (self.get_label(item_dotpath), butn_html))
            return mark_safe('%s<table><tr>%s</tr></table>' % (data_html, '</tr><tr>'.join(rows)))
        else:
            butn_html = self.render_button(dotpath, edit=bool(value))
            desc_html = self.get_label(dotpath, value)
            return mark_safe(''.join((data_html, desc_html, butn_html)))

class DotPathField(HiddenJSONField):
    widget = DotPathWidget
    
    def __init__(self, *args, **kwargs):
        self.dotpath = kwargs.pop('dotpath')
        #if 'widget' not in kwargs or self.widget == kwargs['widget']:
        #    kwargs['widget'] = self.widget(dotpath=self.dotpath)
        super(DotPathField, self).__init__(*args, **kwargs)
        self.widget.dotpath = self.dotpath
    
    def prepare_value(self, value):
        return value


class TypedSchemaFieldWidget(DotPathWidget):
    input_type = 'submit'
    
    def __init__(self, dotpath=None, schema_property=None):
        self.dotpath = dotpath
        self.schema_property = schema_property
        super(DotPathWidget, self).__init__()
    
    def render_type_dropdown(self, dotpath):
        options = list()
        for val, label in self.schema_property.get_schema_choices():
            options.append(u'<option value="%s">%s</option>' % (escape(force_unicode(val)), escape(force_unicode(label))))
        data = {'next_dotpath':dotpath,
                'name':self.schema_property.type_field_name,}
        name = '[fragment-passthrough]%s' % urlencode(data)
        return u'<select name="%s">%s</select>' % (name, '\n'.join(options))
    
    def get_label(self, dotpath, value=None):
        if value:
            return escape(force_unicode(value))
        return self.render_type_dropdown(dotpath)

class TypedSchemaField(DotPathField):
    widget = TypedSchemaFieldWidget
    
    def __init__(self, *args, **kwargs):
        self.schema_property = kwargs.pop('schema_property')
        if hasattr(self.schema_property, 'subfield'):
            self.schema_property = self.schema_property.subfield
        super(TypedSchemaField, self).__init__(*args, **kwargs)
        self.widget.schema_property = self.schema_property




from django.http import HttpResponseRedirect, HttpResponse
from django.utils.encoding import force_unicode
from django.utils.translation import ugettext as _
from django.utils.html import escape, escapejs
from django.views.generic import TemplateView, View

from base import AdminViewMixin
from breadcrumbs import Breadcrumb
import helpers

from dockit import views
from dockit.models import create_temporary_document_class
from dockit.schema.fields import ListField
from dockit.schema.common import UnSet
from dockit.schema.schema import Schema

from urllib import urlencode
from urlparse import parse_qsl

CALL_BACK = "" #TODO

class DocumentViewMixin(AdminViewMixin):
    template_suffix = None
    template_name = None
    form_class = None
    object = None
    
    @property
    def document(self):
        return self.admin.model
    
    @property
    def schema(self):
        return self.admin.schema
    
    def get_template_names(self):
        if self.template_name:
            return [self.template_name]
        opts = self.document._meta
        app_label = opts.app_label
        object_name = opts.object_name.lower()
        return ['admin/%s/%s/%s.html' % (app_label, object_name, self.template_suffix),
                'admin/%s/%s.html' % (app_label, self.template_suffix),
                'admin/%s.html' % self.template_suffix]
    
    #def get_queryset(self):
    #    return self.model.objects.all()
    
    def get_breadcrumbs(self):
        return self.admin.get_base_breadcrumbs()
    
    def get_context_data(self, **kwargs):
        opts = self.document._meta
        obj = self.object
        context = AdminViewMixin.get_context_data(self, **kwargs)
        context.update({'root_path': self.admin_site.root_path,
                        'app_label': opts.app_label,
                        'opts': opts,
                        'module_name': force_unicode(opts.verbose_name_plural),
                        'breadcrumbs': self.get_breadcrumbs(),
                        'has_add_permission': self.admin.has_add_permission(self.request),
                        'has_change_permission': self.admin.has_change_permission(self.request, obj),
                        'has_delete_permission': self.admin.has_delete_permission(self.request, obj),
                        'has_file_field': True, # FIXME - this should check if form or formsets have a FileField,
                        'has_absolute_url': hasattr(self.document, 'get_absolute_url'),
                        #'content_type_id': ContentType.objects.get_for_model(self.model).id,
                        'save_as': self.admin.save_as,
                        'save_on_top': self.admin.save_on_top,})
        return context

class BaseFragmentViewMixin(DocumentViewMixin):
    def is_polymorphic(self, schema):
        return bool(schema._meta.typed_field)
    
    def should_prompt_polymorphic_type(self, schema, obj=None):
        if self.is_polymorphic(schema):
            foo = schema._meta.typed_field
            return obj is None or not getattr(obj, schema._meta.typed_field, None)
        return False
    
    def needs_typed_selection(self, schema, obj=None):
        if schema._meta.typed_field and schema._meta.typed_field in self.request.GET:
            return False
        return self.should_prompt_polymorphic_type(schema, obj)
    
    def dotpath(self):
        return self.request.GET.get('_dotpath', None)
    
    def parent_dotpath(self):
        return self.request.GET.get('_parent_dotpath', None)
    
    def temporary_document_id(self):
        return self.request.GET.get('_tempdoc', None)
    
    def get_temporary_store(self):
        if not hasattr(self, '_temporary_store'):
            temp_doc_id = self.temporary_document_id()
            if temp_doc_id:
                storage = self.temp_document.objects.get(pk=temp_doc_id)
            else:
                if getattr(self, 'object', None):
                    storage = self.temp_document.create_from_instance(self.object)
                else:
                    storage = self.temp_document()
            self._temporary_store = storage
        return self._temporary_store
    
    def get_active_object(self):
        if self.dotpath():
            val = self.get_temporary_store()
            return val.dot_notation_to_value(self.dotpath())
        return self.get_temporary_store()
    
    @property
    def temp_document(self):
        if not hasattr(self, '_temp_document'):
            self._temp_document = create_temporary_document_class(self.document)
        return self._temp_document
    
    def get_breadcrumbs(self):
        breadcrumbs = DocumentViewMixin.get_breadcrumbs(self)
        obj = self.object
        breadcrumbs.append(self.admin.get_instance_breadcrumb(obj))
        return breadcrumbs

from forms import TypeSelectionForm

class SchemaTypeSelectionView(BaseFragmentViewMixin, TemplateView):
    template_name = 'admin/type_selection_form.html'
    form_class = TypeSelectionForm
    schema = None
    
    def get_form_class(self):
        return self.form_class
    
    def get_form_kwargs(self):
        assert self.schema
        return {'schema': self.schema,
                'initial':self.request.GET}
    
    def get_form(self, form_cls):
        return form_cls(**self.get_form_kwargs())
    
    def get_admin_form_kwargs(self):
        return {
            'fieldsets': [(None, {'fields': [self.schema._meta.typed_field]})],
            'prepopulated_fields': dict(),
            'readonly_fields': [],
            'model_admin': self.admin,
        }
    
    def create_admin_form(self):
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        admin_form = helpers.AdminForm(form, **self.get_admin_form_kwargs())
        return admin_form
    
    def get_breadcrumbs(self):
        breadcrumbs = DocumentViewMixin.get_breadcrumbs(self)
        obj = getattr(self, 'object', None)
        breadcrumbs.append(self.admin.get_instance_breadcrumb(obj))
        if self.dotpath():
            breadcrumbs.append(Breadcrumb(self.dotpath()))
        return breadcrumbs
    
    def get_context_data(self, **kwargs):
        context = DocumentViewMixin.get_context_data(self, **kwargs)
        context.update(TemplateView.get_context_data(self, **kwargs))
        context['adminform'] = self.create_admin_form()
        context['form_url'] = self.request.get_full_path()
        return context

class FragmentViewMixin(BaseFragmentViewMixin):
    template_suffix = 'schema_form'
    
    def create_admin_form(self):
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        admin_form = helpers.AdminForm(form, **self.get_admin_form_kwargs())
        return admin_form
    
    def get_admin_form_kwargs(self):
        return {
            'fieldsets': self.get_fieldsets(),
            'prepopulated_fields': self.get_prepopulated_fields(),
            'readonly_fields': self.get_readonly_fields(),
            'model_admin': self.admin,
        }
    
    def get_form_class(self):
        """
        Returns the form class to use in this view
        """
        if self.form_class:
            return self.form_class
        else:
            return self._generate_form_class()
    
    def get_readonly_fields(self):
        return self.admin.get_readonly_fields(self.request)
    
    def get_fieldsets(self, obj=None):
        "Hook for specifying fieldsets for the add form."
        form = self.get_form_class()
        fields = form.base_fields.keys()
        return [(None, {'fields': fields})]
    
    def get_prepopulated_fields(self):
        return self.admin.prepopulated_fields
    
    def get_formsets(self):
        formsets_cls = self.admin.get_formsets(self.request, self.get_active_object())
        
        prefixes = {}
        formsets = list()
        for FormSet, inline in zip(formsets_cls,
                                   self.admin.get_inline_instances()):
                prefix = FormSet.get_default_prefix()
                prefixes[prefix] = prefixes.get(prefix, 0) + 1
                if prefixes[prefix] != 1:
                    prefix = "%s-%s" % (prefix, prefixes[prefix])
                kwargs = self.get_formset_kwargs()
                kwargs['prefix'] = prefix
                if self.dotpath():
                    kwargs['dotpath'] = self.dotpath() + '.' + inline.dotpath
                else:
                    kwargs['dotpath'] = inline.dotpath
                formset = FormSet(**kwargs)
                formsets.append(formset)
        return formsets
    
    def get_formset_kwargs(self):
        kwargs = self.get_form_kwargs()
        kwargs.pop('initial', None)
        return kwargs
    
    def get_inline_admin_formsets(self):
        formsets = self.get_formsets()
        obj = self.get_active_object()
        inline_admin_formsets = []
        for inline, formset in zip(self.admin.get_inline_instances(), formsets):
            fieldsets = list(inline.get_fieldsets(self.request))
            readonly = list(inline.get_readonly_fields(self.request))
            inline_admin_formset = helpers.InlineAdminFormSet(inline, formset,
                fieldsets, readonly, model_admin=self)
            inline_admin_formsets.append(inline_admin_formset)
        
        return inline_admin_formsets
    
    def get_breadcrumbs(self):
        breadcrumbs = BaseFragmentViewMixin.get_breadcrumbs(self)
        if self.dotpath():
            breadcrumbs.append(Breadcrumb(self.dotpath()))
        return breadcrumbs
    
    def get_context_data(self, **kwargs):
        context = AdminViewMixin.get_context_data(self, **kwargs)
        opts = self.schema._meta
        context.update({'title': _('Add %s') % force_unicode(opts.verbose_name),
                        'show_save': True,
                        'show_delete_link': bool(self.dotpath()), #TODO is it a new subobject?
                        'show_save_and_add_another': False,
                        'show_save_and_continue': True,
                        'add': True,
                        'add_another': True,
                        'cancel': False,
                        'change': False,
                        'delete': False,
                        'breadcrumbs': self.get_breadcrumbs(),
                        'dotpath': self.dotpath(),
                        'tempdoc': self.get_temporary_store(),
                        'adminform':self.create_admin_form(),
                        'inline_admin_formsets': self.get_inline_admin_formsets(),})
        context['media'] += context['adminform'].form.media
        for inline in context['inline_admin_formsets']:
            context['media'] += inline.media
            #TODO this does not work as media is a property, fix this and js will be pulled from the inline form
            context['media'] += inline.formset.form.media
        
        obj = None
        if hasattr(self, 'object'):
            obj = self.object
        
        context.update({'root_path': self.admin_site.root_path,
                        'app_label': opts.app_label,
                        'opts': opts,
                        'original': obj,
                        'module_name': force_unicode(opts.verbose_name_plural),
                        'has_add_permission': self.admin.has_add_permission(self.request),
                        'has_change_permission': self.admin.has_change_permission(self.request, obj),
                        'has_delete_permission': self.admin.has_delete_permission(self.request, obj),
                        'has_file_field': True, # FIXME - this should check if form or formsets have a FileField,
                        'has_absolute_url': hasattr(self.document, 'get_absolute_url'),
                        #'content_type_id': ContentType.objects.get_for_model(self.model).id,
                        'save_as': self.admin.save_as,
                        'save_on_top': self.admin.save_on_top,})
        context['object_tools'] = self.get_object_tools(context)
        return context
    
    @property
    def form_class(self):
        return self.admin.form_class
    
    def fragment_info(self):
        token = '[fragment]'
        for key in self.request.POST.iterkeys():
            if key.startswith(token):
                info = dict(parse_qsl(key[len(token):]))
                if info:
                    return info
        return {}
    
    def fragment_passthrough(self):
        dotpath = self.next_dotpath()
        token = '[fragment-passthrough]'
        passthrough = dict()
        for key, value in self.request.POST.iteritems():
            if key.startswith(token):
                info = dict(parse_qsl(key[len(token):]))
                if info and info.pop('next_dotpath', None) == dotpath:
                    passthrough[info['name']] = value
        return passthrough
    
    def next_dotpath(self):
        info = self.fragment_info()
        return info.get('next_dotpath', None)
    
    def get_object_tools(self, context):
        object_tools = list()
        for object_tool in self.admin.get_object_tools(self.request):
            object_tools.append(object_tool.render(self.request, context))
        return object_tools
    
    def formfield_for_field(self, prop, field, **kwargs):
        return self.admin.formfield_for_field(prop, field, self, **kwargs)
    
    def get_base_schema(self):
        return self.admin.schema
    
    def get_schema(self):
        '''
        Retrieves the currently active schema, taking into account dynamic typing
        '''
        return self.get_base_schema()
    
    def _generate_form_class(self):
        form_cls = self.admin.get_form_class(self.request)
        
        class CustomDocumentForm(form_cls):
            class Meta:
                document = self.temp_document
                schema = self.get_schema()
                form_field_callback = self.formfield_for_field
                dotpath = self.dotpath() or None
                exclude = self.admin.get_excludes() + self.get_readonly_fields()
                #TODO fix readonly field behavior
        return CustomDocumentForm
    
    def get_form_kwargs(self, **kwargs):
        kwargs['instance'] = self.get_temporary_store()
        if self.request.method.upper() in ('POST', 'PUT'):
            kwargs['data'] = self.request.POST
            kwargs['files'] = self.request.FILES
        if self.dotpath():
            kwargs['dotpath'] = self.dotpath()
        
        #populate typed field
        schema = self.get_schema()
        if schema._meta.typed_field and schema._meta.typed_field in self.request.GET:
            kwargs.setdefault('initial', {})
            kwargs['initial'][schema._meta.typed_field] = self.request.GET[schema._meta.typed_field]
        return kwargs
    
    def delete_subobject(self):
        temp = self.get_temporary_store()
        params = {}
        
        #TODO in FragmentViewMixin, get_effective_parent_dotpath()
        next_dotpath = self.parent_dotpath()
        if next_dotpath is None:
            dotpath = self.dotpath()
            if '.' in dotpath:
                next_dotpath = dotpath[:dotpath.rfind('.')]
            field = temp.dot_notation_to_field(next_dotpath)
            if isinstance(field, ListField):
                if '.' in next_dotpath:
                    next_dotpath = next_dotpath[:next_dotpath.rfind('.')]
                else:
                    next_dotpath = None
        
        if next_dotpath:
            params['_dotpath'] = next_dotpath
        #TODO make this a blessed function
        temp.dot_notation_set_value(self.dotpath(), UnSet)
        temp.save()
        params['_tempdoc'] = temp.get_id()
        return HttpResponseRedirect('%s?%s' % (self.admin.reverse(self.admin.app_name+'_change', self.kwargs['pk']), urlencode(params)))
    
    def post(self, request, *args, **kwargs):
        if self.dotpath() and "_delete" in self.request.POST:
            return self.delete_subobject()
        
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        formsets = self.get_formsets()
        if not form.is_valid():
            return self.form_invalid(form)#, formsets)
        
        for formset in formsets:
            if not formset.is_valid():
                return self.form_invalid(form)#, formsets)
        
        obj = form.save(commit=False) #CONSIDER this would normally be done in form_valid
        for formset in formsets:
            formset.save(instance=obj)#form.target_object)
        obj.save()
        assert obj._meta.collection == self.temp_document._meta.collection
        
        
        if self.next_dotpath():
            info = self.fragment_info()
            passthrough = self.fragment_passthrough()
            params = {'_dotpath': self.next_dotpath(),
                      '_parent_dotpath': self.dotpath() or '',
                      '_tempdoc': obj.get_id(),}
            params.update(passthrough)
            return HttpResponseRedirect('%s?%s' % (request.path, urlencode(params)))
        if self.dotpath():
            params = {'_tempdoc':obj.get_id(),}
            
            #if they signaled to continue editing
            if self.request.POST.get('_continue', False):
                next_dotpath = self.dotpath()
            else:
                next_dotpath = self.parent_dotpath()
                if next_dotpath is None:
                    dotpath = self.dotpath()
                    if '.' in dotpath:
                        next_dotpath = dotpath[:dotpath.rfind('.')]
                    field = obj.dot_notation_to_field(next_dotpath)
                    if isinstance(field, ListField):
                        if '.' in next_dotpath:
                            next_dotpath = next_dotpath[:next_dotpath.rfind('.')]
                        else:
                            next_dotpath = None
            if next_dotpath:
                params['_dotpath'] = next_dotpath
            return HttpResponseRedirect('%s?%s' % (request.path, urlencode(params)))
        
        #now to create the object!
        if obj._meta.collection != self.document._meta.collection:
            self.object = obj.commit_changes(self.kwargs.get('pk', None))
        else:
            self.object = obj
        if 'pk' in self.kwargs:
            assert str(self.object.pk) == self.kwargs['pk']
        if self.temporary_document_id():
            self.get_temporary_store().delete()
        return self.form_valid(form)
    
    def form_valid(self, form):
        if "_popup" in self.request.POST:
            return HttpResponse(CALL_BACK % \
                # escape() calls force_unicode.
                {'value': escape(self.object.get_id()), 
                 'name': escapejs(self.document._meta.verbose_name)})
        if '_continue' in self.request.POST:
            return HttpResponseRedirect(self.admin.reverse(self.admin.app_name+'_change', self.object.get_id()))
        if '_addanother' in self.request.POST:
            return HttpResponseRedirect(self.admin.reverse(self.admin.app_name+'_add'))
        return HttpResponseRedirect(self.admin.reverse(self.admin.app_name+'_changelist'))

class IndexView(DocumentViewMixin, views.ListView):
    template_suffix = 'change_list'
    
    def get_changelist(self):
        if not hasattr(self, 'changelist'):
            changelist_cls = self.admin.get_changelist(self.request)
            self.changelist = changelist_cls(request=self.request,
                                        model=self.document,
                                        list_display=self.admin.list_display,
                                        list_display_links=self.admin.list_display_links,
                                        list_filter=self.admin.list_filter,
                                        date_hierarchy=self.admin.date_hierarchy,
                                        search_fields=self.admin.search_fields,
                                        list_select_related=self.admin.list_select_related,
                                        list_per_page=self.admin.list_per_page,
                                        list_editable=self.admin.list_editable,
                                        model_admin=self.admin,)
        return self.changelist
    
    def get_context_data(self, **kwargs):
        context = views.ListView.get_context_data(self, **kwargs)
        context.update(DocumentViewMixin.get_context_data(self, **kwargs))
        cl = self.get_changelist()
        context.update({'cl': cl,
                        'title': cl.title,
                        'is_popup': cl.is_popup,})
        
        return context
    
    def get_queryset(self):
        cl = self.get_changelist()
        return cl.get_query_set()

class ListFieldIndexView(BaseFragmentViewMixin, views.DetailView):
    template_suffix = 'listfield_change_list'
    
    def get_changelist_class(self):
        from changelist import ListFieldChangeList
        return ListFieldChangeList
    
    def get_changelist(self):
        if not hasattr(self, 'changelist'):
            instance = self.get_temporary_store()
            dotpath = self.dotpath()
            changelist_cls = self.get_changelist_class()
            #CONSIDER: schemaadmin should have these values
            self.changelist = changelist_cls(request=self.request,
                                        model=self.schema,
                                        instance=instance,
                                        dotpath=dotpath,
                                        list_display=self.admin.list_display,
                                        list_display_links=self.admin.list_display_links,
                                        list_filter=self.admin.list_filter,
                                        date_hierarchy=self.admin.date_hierarchy,
                                        search_fields=self.admin.search_fields,
                                        list_select_related=self.admin.list_select_related,
                                        list_per_page=self.admin.list_per_page,
                                        list_editable=self.admin.list_editable,
                                        model_admin=self.admin,)
        return self.changelist
    
    def get_breadcrumbs(self):
        breadcrumbs = BaseFragmentViewMixin.get_breadcrumbs(self)
        breadcrumbs.append(Breadcrumb(self.dotpath()))
        return breadcrumbs
    
    def get_context_data(self, **kwargs):
        context = views.DetailView.get_context_data(self, **kwargs)
        context.update(BaseFragmentViewMixin.get_context_data(self, **kwargs))
        cl = self.get_changelist()
        context.update({'cl': cl,
                        'title': cl.title,
                        'is_popup': cl.is_popup,})
        context['object_list'] = cl.get_query_set()
        
        params = self.request.GET.copy()
        params['_dotpath'] = '%s.%s' % (params['_dotpath'], len(context['object_list']))
        context['add_link'] = './?%s' % params.urlencode()
        
        return_dotpath = self.dotpath()
        if '.' in return_dotpath:
            return_dotpath = return_dotpath.rsplit('.', 1)[0]
        else:
            return_dotpath = ''
        params['_dotpath'] = return_dotpath
        context['return_link'] = './?%s' % params.urlencode()
        return context

class DocumentProxyView(BaseFragmentViewMixin, View):
    def dispatch(self, request, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.request = request
        self.object = self.get_object()
        schema = self.get_base_schema()
        
        if not self.dotpath() and self.needs_typed_selection(schema, self.get_temporary_store()):
            admin = self.admin.create_admin_for_schema(schema)
            return admin.get_select_schema_view(object=self.object)(request, *args, **kwargs)
        
        schema = self.get_schema()
        admin = self.admin.create_admin_for_schema(schema)
        field = self.get_field()
        
        #detect list field
        if field and isinstance(field, ListField):
            return admin.get_field_list_index_view(object=self.object)(request, *args, **kwargs)
        
        if 'pk' in kwargs:
            return admin.get_update_view(object=self.object)(request, *args, **kwargs)
        else:
            return admin.get_create_view(object=self.object)(request, *args, **kwargs)
    
    def get_base_schema(self):
        '''
        Retrieves base schema, taking into account dynamic typing
        '''
        schema = self.admin.model
        if schema._meta.typed_field:
            field = schema._meta.fields[schema._meta.typed_field]
            if schema._meta.typed_field in self.request.GET:
                key = self.request.GET[schema._meta.typed_field]
                try:
                    schema = field.schemas[key]
                except KeyError:
                    #TODO emit a warning
                    pass
            else:
                if self.temporary_document_id():
                    obj = self.get_temporary_store()
                else:
                    obj = self.object
                if obj:
                    try:
                        schema = field.schemas[obj[schema._meta.typed_field]]
                    except KeyError:
                        #TODO emit a warning
                        pass
            #KeyErrors cause the base schema to return, this should cause needs_typed_selection to return true
        return schema
    
    def get_field(self):
        return self.admin.get_field(self.get_base_schema(), self.dotpath(), self.get_temporary_store())
    
    def get_schema(self):
        '''
        Retrieves the currently active schema, taking into account dynamic typing
        '''
        schema = None
        
        if self.dotpath():
            field = self.get_field()
            if getattr(field, 'subfield', None):
                field = field.subfield
            if getattr(field, 'schema'):
                schema = field.schema
                if schema._meta.typed_field:
                    typed_field = schema._meta.fields[schema._meta.typed_field]
                    if schema._meta.typed_field in self.request.GET:
                        key = self.request.GET[schema._meta.typed_field]
                        schema = typed_field.schemas[key]
                    else:
                        obj = self.get_active_object()
                        if obj is not None and isinstance(obj, Schema):
                            schema = type(obj)
            else:
                assert False
        else:
            schema = self.get_base_schema()
        assert issubclass(schema, Schema)
        return schema
    
    def get_object(self):
        if 'pk' in self.kwargs:
            return self.document.objects.get(pk=self.kwargs['pk'])
        return None

class CreateView(FragmentViewMixin, views.CreateView):
    def dispatch(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs
        schema = self.get_schema()
        obj = self.get_active_object()
        if self.needs_typed_selection(schema, obj):
            return self.admin.get_select_schema_view()(request, *args, **kwargs)
        return super(CreateView, self).dispatch(request, *args, **kwargs)
    
    def get_context_data(self, **kwargs):
        context = views.CreateView.get_context_data(self, **kwargs)
        context.update(FragmentViewMixin.get_context_data(self, **kwargs))
        opts = self.schema._meta
        context.update({'title': _('Add %s') % force_unicode(opts.verbose_name),
                        'show_delete': False,
                        'add': True,
                        'change': False,
                        'delete': False,
                        'adminform':self.create_admin_form(),})
        context['media'] += context['adminform'].form.media
        return context
    
    def form_valid(self, form):
        self.admin.log_addition(self.request, self.object)
        return FragmentViewMixin.form_valid(self, form)
    
class UpdateView(FragmentViewMixin, views.UpdateView):
    def dispatch(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs
        if not self.object:
            self.object = self.get_object()
        schema = self.get_schema()
        obj = self.get_active_object()
        if self.needs_typed_selection(schema, obj):
            return self.admin.get_select_schema_view(object=self.object)(request, *args, **kwargs)
        return super(UpdateView, self).dispatch(request, *args, **kwargs)
    
    def get_context_data(self, **kwargs):
        context = views.UpdateView.get_context_data(self, **kwargs)
        context.update(FragmentViewMixin.get_context_data(self, **kwargs))
        
        obj = self.object
        opts = self.schema._meta
        context.update({'title':_('Change %s') % force_unicode(opts.verbose_name),
                        'object_id': obj.get_id,
                        'original': obj,
                        'change': True,
                        'add': False,
                        'delete': False,
                        'adminform':self.create_admin_form(),})
        context['media'] += context['adminform'].form.media
        return context
    
    def form_valid(self, form):
        self.admin.log_change(self.request, self.object, '')
        return FragmentViewMixin.form_valid(self, form)

class DeleteView(DocumentViewMixin, views.DetailView):
    template_suffix = 'delete_selected_confirmation'
    title = _('Delete')
    key = 'delete'
    
    def get_context_data(self, **kwargs):
        context = views.DetailView.get_context_data(self, **kwargs)
        #context.update(FragmentViewMixin.get_context_data(self, **kwargs))
        #TODO add what will be deleted
        return context
    
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        object_repr = unicode(self.object)
        self.object.delete()
        self.admin.log_deletion(request, self.object, object_repr)
        return HttpResponseRedirect(self.admin.reverse(self.admin.app_name+'_changelist'))

class HistoryView(DocumentViewMixin, views.ListView):
    title = _('History')
    key = 'history'




from dockit.models import DockitPermission
from dockit.schema.signals import document_registered
from dockit.schema.common import COLLECTIONS

from django.contrib.auth import models as auth_app
from django.contrib.contenttypes.models import ContentType

def _get_permission_codename(action, opts):
    return u'%s.%s' % (opts.collection, action)

def _get_all_permissions(opts):
    "Returns (codename, name) for all permissions in the given opts."
    perms = []
    for action in ('add', 'change', 'delete'):
        perms.append((_get_permission_codename(action, opts), u'Can %s %s' % (action, opts.verbose_name_raw)))
    return perms + opts.permissions

#code => <collection>.code
#lookup code => dockit.<collection>.code
#content type = DockitPermission

#user.has_perm(dockit.somecollection.code)

def create_permissions(documents, verbosity, **kwargs):
    # This will hold the permissions we're looking for as
    # (content_type, (codename, name))
    searched_perms = list()
    ctype = ContentType.objects.get_for_model(DockitPermission)
    # The codenames and ctypes that should exist.
    for klass in documents:
        for perm in _get_all_permissions(klass._meta):
            searched_perms.append(perm)

    # Find all the Permissions that have a context_type for a model we're
    # looking for.  We don't need to check for codenames since we already have
    # a list of the ones we're going to create.
    all_perms = set(auth_app.Permission.objects.filter(
        content_type=ctype,
    ).values_list(
        "codename", flat=True
    ))
    
    for codename, name in searched_perms:
        # If the permissions exists, move on.
        if codename in all_perms:
            continue
        p = auth_app.Permission.objects.create(
            codename=codename,
            name=name,
            content_type=ctype
        )
        if verbosity >= 2:
            print "Adding permission '%s'" % p

def on_document_registered(document, **kwargs):
    create_permissions([document], 1)
document_registered.connect(on_document_registered)

create_permissions(COLLECTIONS.values(), 1)





from django.forms.formsets import BaseFormSet, formset_factory
from django.utils.translation import ugettext_lazy as _, ugettext

from forms import DocumentForm, documentform_factory

class BaseDocumentFormSet(BaseFormSet):
    """
    A ``FormSet`` for editing a queryset and/or adding new objects to it.
    """
    document = None

    def __init__(self, data=None, files=None, auto_id='id_%s', prefix=None,
                 queryset=None, **kwargs):
        self.queryset = queryset
        defaults = {'data': data, 'files': files, 'auto_id': auto_id, 'prefix': prefix}
        defaults.update(kwargs)
        super(BaseDocumentFormSet, self).__init__(**defaults)

    def initial_form_count(self):
        """Returns the number of forms that are required in this FormSet."""
        if not (self.data or self.files):
            return len(self.get_queryset())
        return super(BaseDocumentFormSet, self).initial_form_count()

    def _existing_object(self, pk):
        if not hasattr(self, '_object_dict'):
            self._object_dict = dict([(i, o) for i, o in enumerate(self.get_queryset())])
        return self._object_dict.get(pk)

    def _construct_form(self, i, **kwargs):
        if self.is_bound and i < self.initial_form_count():
            # Import goes here instead of module-level because importing
            # django.db has side effects.
            
            kwargs['instance'] = self._existing_object(i)
        if i < self.initial_form_count() and not kwargs.get('instance'):
            kwargs['instance'] = self.get_queryset()[i]
        return super(BaseDocumentFormSet, self)._construct_form(i, **kwargs)

    def get_queryset(self):
        if not hasattr(self, '_queryset'):
            if self.queryset is not None:
                qs = self.queryset
            else:
                qs = self.document.objects.all()

            # If the queryset isn't already ordered we need to add an
            # artificial ordering here to make sure that all formsets
            # constructed from this queryset have the same form order.
            #if not qs.ordered:
            #    qs = qs.order_by(self.model._meta.pk.name)

            # Removed queryset limiting here. As per discussion re: #13023
            # on django-dev, max_num should not prevent existing
            # related objects/inlines from being displayed.
            self._queryset = qs
        return self._queryset

    def save_new(self, form, commit=True):
        """Saves and returns a new model instance for the given form."""
        return form.save(commit=commit)

    def save_existing(self, form, instance, commit=True):
        """Saves and returns an existing model instance for the given form."""
        return form.save(commit=commit)

    def save(self, commit=True):
        """Saves model instances for every form, adding and changing instances
        as necessary, and returns the list of instances.
        """
        if not commit:
            self.saved_forms = []
            def save_m2m():
                for form in self.saved_forms:
                    form.save_m2m()
            self.save_m2m = save_m2m
        return self.save_existing_objects(commit) + self.save_new_objects(commit)

    def clean(self):
        pass

    def get_form_error(self):
        return ugettext("Please correct the duplicate values below.")

    def save_existing_objects(self, commit=True):
        self.changed_objects = []
        self.deleted_objects = []
        if not self.get_queryset():
            return []

        saved_instances = []
        for i, form in enumerate(self.initial_forms):
            obj = self._existing_object(i)
            if self.can_delete and self._should_delete_form(form):
                self.deleted_objects.append(obj)
                #obj.delete()
                continue
            if form.has_changed():
                self.changed_objects.append((obj, form.changed_data))
            saved_instances.append(self.save_existing(form, obj, commit=commit))
            if not commit:
                self.saved_forms.append(form)
        return saved_instances

    def save_new_objects(self, commit=True):
        self.new_objects = []
        for form in self.extra_forms:
            if not form.has_changed():
                continue
            # If someone has marked an add form for deletion, don't save the
            # object.
            if self.can_delete and self._should_delete_form(form):
                continue
            self.new_objects.append(self.save_new(form, commit=commit))
            if not commit:
                self.saved_forms.append(form)
        return self.new_objects

    def add_fields(self, form, index):
        #CONSIDER there is no parent object, more like a document/instance and a dotpath
        """Add a hidden field for the object's primary key."""
        #TODO
        super(BaseDocumentFormSet, self).add_fields(form, index)

def documentformset_factory(document, form=DocumentForm, formfield_callback=None,
                         formset=BaseDocumentFormSet,
                         dotpath=None, schema=None,
                         extra=1, can_delete=False, can_order=False,
                         max_num=None, fields=None, exclude=None):
    """
    Returns a FormSet class for the given Document class.
    """
    form = documentform_factory(document, form=form, fields=fields, exclude=exclude,
                             dotpath=dotpath, schema=schema,
                             formfield_callback=formfield_callback)
    FormSet = formset_factory(form, formset, extra=extra, max_num=max_num,
                              can_order=can_order, can_delete=can_delete)
    FormSet.document = document
    return FormSet


# InlineFormSets #############################################################

class BaseInlineFormSet(BaseDocumentFormSet): #simply merge as one?
    """A formset for child objects related to a parent."""
    def __init__(self, data=None, files=None, instance=None,
                 save_as_new=False, prefix=None, dotpath=None):
        self.instance = instance
        self.save_as_new = save_as_new
        self.dotpath = dotpath or self.form._meta.dotpath
        self.singleton = False
        # is there a better way to get the object descriptor?
        qs = self.instance.dot_notation(self.base_dotpath) or []
        if not isinstance(qs, (list, set)):
            qs = [qs]
            self.singleton = True
        super(BaseInlineFormSet, self).__init__(data, files, prefix=prefix,
                                                queryset=qs)
    
    @property
    def base_dotpath(self):
        return self.dotpath.rsplit('.', 1)[0]
    
    def initial_form_count(self):
        if self.save_as_new:
            return 0
        return super(BaseInlineFormSet, self).initial_form_count()

    def _construct_form(self, i, **kwargs):
        if self.singleton:
            kwargs['dotpath'] = self.base_dotpath
        else:
            kwargs['dotpath'] = '%s.%i' % (self.base_dotpath, i)
        kwargs['instance'] = self.instance
        form = super(BaseDocumentFormSet, self)._construct_form(i, **kwargs)
        if self.save_as_new:
            # Remove the primary key from the form's data, we are only
            # creating new instances
            #form.data[form.add_prefix(self._pk_field.name)] = None

            # Remove the foreign key from the form's data
            #form.data[form.add_prefix(self.fk.name)] = None
            pass

        # Set the fk value here so that the form can do it's validation.
        #?????
        #setattr(form.instance, self.fk.get_attname(), self.instance.pk)
        return form

    #def get_default_prefix(self):
    #    return self.dotpath.rsplit('.', 1)[-1]
    #get_default_prefix = classmethod(get_default_prefix)
    
    def save_new(self, form, commit=True):
        """Saves and returns a new model instance for the given form."""
        return form._inner_save()

    def save_existing(self, form, instance, commit=True):
        """Saves and returns an existing model instance for the given form."""
        return form._inner_save()
    
    def save(self, commit=True, instance=None):
        """Saves model instances for every form, adding and changing instances
        as necessary, and returns the list of instances.
        """
        if instance:
            self.instance = instance
        if not commit:
            self.saved_forms = []
            def save_m2m():
                for form in self.saved_forms:
                    form.save_m2m()
            self.save_m2m = save_m2m
        new_list = self.save_existing_objects(commit) + self.save_new_objects(commit)
        if commit:
            if self.singleton:
                self.instance.dot_notation_set_value(self.base_dotpath, new_list[0])
            else:
                self.instance.dot_notation_set_value(self.base_dotpath, new_list)
        return new_list


def inlinedocumentformset_factory(document, dotpath, form=DocumentForm,
                          formset=BaseInlineFormSet,
                          fields=None, exclude=None, schema=None,
                          extra=3, can_order=False, can_delete=True, max_num=None,
                          formfield_callback=None):
    """
    Returns an ``InlineFormSet`` for the given kwargs.

    You must provide ``fk_name`` if ``model`` has more than one ``ForeignKey``
    to ``parent_model``.
    """
    kwargs = {
        'form': form,
        'formfield_callback': formfield_callback,
        'formset': formset,
        'extra': extra,
        'can_delete': can_delete,
        'can_order': can_order,
        'fields': fields,
        'exclude': exclude,
        'max_num': max_num,
        'schema': schema,
        'dotpath': dotpath + '.*',
    }
    FormSet = documentformset_factory(document, **kwargs)
    #FormSet.fk = fk
    return FormSet




from dockit.backends.indexer import BaseIndexer

from backend import MongoDocumentStorage

from pymongo.objectid import ObjectId

class ExactIndexer(BaseIndexer):
    def _get_key_value(self):
        dotpath = self.filter_operation.dotpath()
        value = self.filter_operation.value
        if dotpath in ('pk', '_pk', '_id') and value:
            value = ObjectId(value)
        if dotpath in ('pk', '_pk'):
            dotpath = '_id'
        return dotpath, value
    
    def filter(self):
        dotpath, value = self._get_key_value()
        return {dotpath: value}
    
    def values(self):
        dotpath, value = self._get_key_value()
        return {dotpath: value}

MongoDocumentStorage.register_indexer(ExactIndexer, 'exact')



class BaseDocumentQuery(object):
    """
    Implemented by the backend to execute a certain index
    """
    def __init__(self, query_index):
        self.query_index = query_index
    
    @property
    def document(self):
        return self.query_index.document
    
    @property
    def backend(self):
        return self.document._meta.get_backend()
    
    def _get_indexer_for_operation(self, document, op):
        return self.backend._get_indexer_for_operation(document, op)
    
    def __len__(self):
        raise NotImplementedError
    
    def count(self):
        return self.__len__()
    
    def exists(self):
        return bool(self.__len__())
    
    def delete(self):
        raise NotImplementedError
    
    def get(self, **kwargs):
        filter_operations = self.query_index._parse_kwargs(kwargs)
        return self.get_from_filter_operations(filter_operations)
    
    def get_from_filter_operations(self, filter_operations):
        raise NotImplementedError
    
    def values(self):
        raise NotImplementedError
    
    def __getitem__(self, val):
        raise NotImplementedError
    
    def __nonzero__(self):
        raise NotImplementedError

class QuerySet(object):
    '''
    Acts as the queryset level caching layer
    '''
    def __init__(self, query):
        self.query = query
    
    def __len__(self):
        #TODO cache
        return self.query.__len__()
    
    def count(self):
        return self.__len__()
    
    def delete(self):
        return self.query.delete()
    
    def values(self):
        return self.query.values()
    
    def get(self, **kwargs):
        #TODO cache
        return self.query.get(**kwargs)
    
    def exists(self):
        return self.query.exists()
    
    def __getitem__(self, val):
        #TODO cache
        return self.query.__getitem__(val)
    
    def __nonzero__(self):
        #TODO cache
        return self.query.__nonzero__()
    
    def __iter__(self):
        return iter(self.query)




from dockit.paginator import Paginator

from django.core.exceptions import ImproperlyConfigured
from django.views.generic import list as listview

class MultipleObjectMixin(listview.MultipleObjectMixin):
    paginator_class = Paginator
    document = None
    
    def get_queryset(self):
        """
        Get the list of items for this view. This must be an interable, and may
        be a queryset (in which qs-specific behavior will be enabled).
        """
        if self.queryset is not None:
            queryset = self.queryset
            if hasattr(queryset, '_clone'):
                queryset = queryset._clone()
        elif self.document is not None:
            queryset = self.document.objects.all()
        else:
            raise ImproperlyConfigured(u"'%s' must define 'queryset' or 'model'"
                                       % self.__class__.__name__)
        return queryset

class BaseListView(MultipleObjectMixin, listview.BaseListView):
    pass

class ListView(listview.MultipleObjectTemplateResponseMixin, BaseListView):
    pass



from django.contrib.admin.views.main import ChangeList as BaseChangeList, MAX_SHOW_ALL_ALLOWED, InvalidPage, IncorrectLookupParameters

class ChangeList(BaseChangeList):
    formset = None
    
    def get_query_set(self):
        return self.root_query_set
    
    def get_results(self, request):
        paginator = self.model_admin.get_paginator(request, self.query_set, self.list_per_page)
        # Get the number of objects, with admin filters applied.
        result_count = paginator.count
        
        # Get the total number of objects, with no admin filters applied.
        # Perform a slight optimization: Check to see whether any filters were
        # given. If not, use paginator.hits to calculate the number of objects,
        # because we've already done paginator.hits and the value is cached.
        full_result_count = result_count
        
        can_show_all = result_count <= MAX_SHOW_ALL_ALLOWED
        multi_page = result_count > self.list_per_page

        # Get the list of objects to display on this page.
        if (self.show_all and can_show_all) or not multi_page:
            result_list = self.query_set
        else:
            try:
                result_list = paginator.page(self.page_num+1).object_list
            except InvalidPage:
                raise IncorrectLookupParameters

        self.result_count = result_count
        self.full_result_count = full_result_count
        if callable(result_list):
            result_list = result_list()
        self.result_list = result_list
        self.can_show_all = can_show_all
        self.multi_page = multi_page
        self.paginator = paginator

class ListFieldChangeList(ChangeList):
    def __init__(self, **kwargs):
        self.instance = kwargs.pop('instance')
        self.dotpath = kwargs.pop('dotpath')
        self.request = kwargs['request']
        super(ListFieldChangeList, self).__init__(**kwargs)
    
    def get_query_set(self):
        def _patch_instance(instance, index): #complete hack
            def serializable_value(*args):
                return index
            instance.serializable_value = serializable_value
            return instance
        return [_patch_instance(obj, index) for index, obj in enumerate(self.instance.dot_notation_to_value(self.dotpath) or [])]
    
    def url_for_result(self, result):
        dotpath = '%s.%s' % (self.dotpath, result.serializable_value())
        params = self.request.GET.copy()
        params['_dotpath'] = dotpath
        return "./?%s" % params.urlencode()




#from django.views.generic.dates import (ArchiveIndexView, YearArchiveView, MonthArchiveView,
#                                     WeekArchiveView, DayArchiveView, TodayArchiveView,
#                                     DateDetailView)
from detail import DetailView
from edit import DocumentFormMixin, CreateView, UpdateView, DeleteView #FormView
from list import ListView




from django.conf.urls.defaults import patterns, url
from django.utils.functional import update_wrapper
from django.utils.encoding import force_unicode
from django.core.urlresolvers import reverse
from django.contrib.admin import widgets
from django import forms

from dockit.paginator import Paginator
from dockit.forms import DocumentForm
from dockit.forms.fields import PrimitiveListField

import views
from widgets import AdminPrimitiveListWidget
from breadcrumbs import Breadcrumb

FORMFIELD_FOR_FIELD_DEFAULTS = {
    forms.DateTimeField: {
        'form_class': forms.SplitDateTimeField,
        'widget': widgets.AdminSplitDateTime
    },
    forms.DateField:       {'widget': widgets.AdminDateWidget},
    forms.TimeField:       {'widget': widgets.AdminTimeWidget},
    #models.TextField:       {'widget': widgets.AdminTextareaWidget},
    #models.URLField:        {'widget': widgets.AdminURLFieldWidget},
    forms.IntegerField:    {'widget': widgets.AdminIntegerFieldWidget},
    #models.BigIntegerField: {'widget': widgets.AdminIntegerFieldWidget},
    forms.CharField:       {'widget': widgets.AdminTextInputWidget},
    forms.ImageField:      {'widget': widgets.AdminFileWidget},
    forms.FileField:       {'widget': widgets.AdminFileWidget},
}

class SchemaAdmin(object):
    #class based views
    create = views.CreateView
    update = views.UpdateView
    delete = views.DeleteView
    select_schema = views.SchemaTypeSelectionView
    field_list_index = views.ListFieldIndexView
    
    raw_id_fields = ()
    fields = None
    exclude = []
    fieldsets = None
    form = forms.ModelForm #only for legacy purposes, remove this and django admin complains
    form_class = None
    filter_vertical = ()
    filter_horizontal = ()
    radio_fields = {}
    prepopulated_fields = {}
    formfield_overrides = {}
    readonly_fields = ()
    declared_fieldsets = None
    
    save_as = False
    save_on_top = False
    paginator = Paginator
    inlines = []
    
    #list display options
    list_display = ('__str__',)
    list_display_links = ()
    list_filter = ()
    list_select_related = False
    list_per_page = 100
    list_editable = ()
    search_fields = ()
    date_hierarchy = None
    ordering = None

    # Custom templates (designed to be over-ridden in subclasses)
    add_form_template = None
    change_form_template = None
    change_list_template = None
    delete_confirmation_template = None
    delete_selected_confirmation_template = None
    object_history_template = None
    
    schema = None

    def __init__(self, model, admin_site, schema=None, documentadmin=None):
        self.model = model
        self.admin_site = admin_site
        self.app_name = model._meta.app_label +'_'+ model._meta.module_name
        overrides = FORMFIELD_FOR_FIELD_DEFAULTS.copy()
        overrides.update(self.formfield_overrides)
        self.formfield_overrides = overrides
        self.schema = schema
        self.documentadmin = documentadmin
        
    def get_view_kwargs(self):
        return {'admin':self,
                'admin_site':self.admin_site,}
    
    def _media(self):
        from django.conf import settings
        from django import forms

        js = ['js/core.js', 'js/admin/RelatedObjectLookups.js',
              'js/jquery.min.js', 'js/jquery.init.js']
        #if self.actions is not None:
        #    js.extend(['js/actions.min.js'])
        if self.prepopulated_fields:
            js.append('js/urlify.js')
            js.append('js/prepopulate.min.js')
        #if self.opts.get_ordered_objects():
        #    js.extend(['js/getElementsBySelector.js', 'js/dom-drag.js' , 'js/admin/ordering.js'])

        return forms.Media(js=['%s%s' % (settings.ADMIN_MEDIA_PREFIX, url_s) for url_s in js])
    media = property(_media)
    
    def has_add_permission(self, request):
        return True
    
    def has_change_permission(self, request, obj=None):
        return True
    
    def has_delete_permission(self, request, obj=None):
        return True
    
    def as_view(self, view, cacheable=False):
        return self.admin_site.admin_view(view, cacheable)
    
    def get_create_view(self, **kwargs):
        params = self.get_view_kwargs()
        params.update(kwargs)
        return self.as_view(self.create.as_view(**params))
    
    def get_update_view(self, **kwargs):
        params = self.get_view_kwargs()
        params.update(kwargs)
        return self.as_view(self.update.as_view(**params))
    
    def get_select_schema_view(self, **kwargs):
        params = self.get_view_kwargs()
        params['schema'] = self.schema
        params.update(kwargs)
        return self.as_view(self.select_schema.as_view(**params))
    
    def get_field_list_index_view(self, **kwargs):
        params = self.get_view_kwargs()
        params.update(kwargs)
        return self.as_view(self.field_list_index.as_view(**params))
    
    def get_model_perms(self, request):
        return {
            'add': self.has_add_permission(request),
            'change': self.has_change_permission(request),
            'delete': self.has_delete_permission(request),
        }
    
    def queryset(self, request):
        return self.model.objects.all()
    
    def get_form_class(self, request, obj=None):
        """
        Returns the form class to use in this view
        """
        if self.form_class:
            return self.form_class
        else:
            return DocumentForm
    
    def _get_schema_fields(self):
        for field in self.schema._meta.fields.itervalues():
            if getattr(field, 'schema', None):
                yield (field, field.schema, False)
            elif getattr(field, 'subfield', None) and getattr(field.subfield, 'schema', None):
                yield (field, field.subfield.schema, True)
    
    def _get_static_schema_fields(self):
        fields = list()
        for field, schema, many in self._get_schema_fields():
            if not schema._meta.is_dynamic():
                fields.append((field, schema, many))
        return fields
    
    def get_excludes(self):
        excludes = set(self.exclude)
        excludes.update([inline.dotpath for inline in self.inlines])
        excludes.update([field.name for field, schema, many in self._get_static_schema_fields()])
        return list(excludes)
    
    def get_default_inline_instances(self, exclude=[]):
        inline_instances = list()
        from inlines import StackedInline
        for field, schema, many in self._get_static_schema_fields():
            if field.name in exclude:
                continue
            kwargs = {'dotpath':field.name + '.*'}
            if not many:
                kwargs['max_num'] = 1
                kwargs['dotpath'] = field.name
            inline_instance = StackedInline(self.model, self.admin_site, schema, self.documentadmin, **kwargs)
            inline_instances.append(inline_instance)
        return inline_instances
    
    def get_inline_instances(self):
        inline_instances = list()
        seen = set()
        for inline_class in self.inlines:
            inline_instance = inline_class(self.model, self.admin_site, None, self.documentadmin)
            inline_instances.append(inline_instance)
            seen.add(inline_instance.dotpath)
        
        inline_instances.extend(self.get_default_inline_instances(exclude=seen))
        
        return inline_instances
    
    def get_formsets(self, request, obj=None):
        inline_instances = self.get_inline_instances()
        for inline in inline_instances:
            yield inline.get_formset(request, obj)
    
    def get_fieldsets(self, request, obj=None):
        "Hook for specifying fieldsets for the add form."
        if self.declared_fieldsets:
            return self.declared_fieldsets
        #form = self.get_form(request, obj)
        #fields = form.base_fields.keys() + list(self.get_readonly_fields(request, obj))
        fields = list()
        for key, field in self.schema._meta.fields.iteritems():
            fields.append(key) #TODO handle exclude
        return [(None, {'fields': fields})]
    
    def formfield_for_field(self, prop, field, view, **kwargs):
        from dockit import schema
        from fields import DotPathField
        from dockit.forms.fields import HiddenJSONField
        request = kwargs.pop('request', None)
        if ((isinstance(prop, schema.ListField) and isinstance(prop.subfield, schema.TypedSchemaField)) or
             isinstance(prop, schema.TypedSchemaField)):
            from fields import TypedSchemaField
            field = TypedSchemaField
            kwargs['dotpath'] = view.dotpath()
            kwargs['schema_property'] = prop
            if self.next_dotpath():
                kwargs['required'] = False
            return field(**kwargs)
        if issubclass(field, HiddenJSONField):
            field = DotPathField
            kwargs['dotpath'] = view.dotpath()
            if view.next_dotpath():
                kwargs['required'] = False
            return field(**kwargs)
        if issubclass(field, PrimitiveListField) and 'subfield' in kwargs:
            subfield_kwargs = dict(kwargs)
            subfield_kwargs.pop('initial', None)
            subfield = self.formfield_for_field(prop, type(subfield_kwargs.pop('subfield')), view, **subfield_kwargs)
            kwargs['subfield'] = subfield
            kwargs['widget'] = AdminPrimitiveListWidget
        if field in self.formfield_overrides:
            opts = dict(self.formfield_overrides[field])
            field = opts.pop('form_class', field)
            kwargs = dict(opts, **kwargs)
        return field(**kwargs)
    
    def log_addition(self, request, object):
        return self.documentadmin.log_addition(request, object)
    
    def log_change(self, request, object, message):
        return self.documentadmin.log_change(request, object, message)
    
    def log_deletion(self, request, object, object_repr):
        return self.documentadmin.log_deletion(request, object, object_repr)
    
    def reverse(self, name, *args, **kwargs):
        return self.documentadmin.reverse(name, *args, **kwargs)
    
    def get_readonly_fields(self, request, schema=None):
        schema = schema or self.schema
        read_only = list()
        for key, field in schema._meta.fields.iteritems():
            if not field.editable:
                read_only.append(key)
        return read_only
    
    def get_paginator(self, request, query_set, paginate_by):
        return self.paginator(query_set, paginate_by)
    
    def get_object_tools(self, request, object=None):
        #TODO object tools are renderable object that are displayed at the top the admin
        #TODO return history object tool if there is an object
        return []
    
    def get_field(self, schema, dotpath, obj=None):
        field = None
        if dotpath and obj:
            field = obj.dot_notation_to_field(dotpath)
            if field is None:
                parent_path = dotpath.rsplit('.', 1)[0]
                print 'no field', dotpath, obj
                
                from dockit.schema.common import DotPathTraverser
                traverser = DotPathTraverser(parent_path)
                traverser.resolve_for_instance(obj)
                info = traverser.resolved_paths
                subschema = info[2]['field'].schema
                fields = subschema._meta.fields
                
                field = obj.dot_notation_to_field(parent_path)
                data = obj._primitive_data
                assert field
        return field
    
    def get_base_breadcrumbs(self):
        return self.documentadmin.get_base_breadcrumbs()
    
    def get_instance_breadcrumb(self, obj=None):
        return self.documentadmin.get_instance_breadcrumb(obj)

class DocumentAdmin(SchemaAdmin):
    # Actions
    actions = []
    #action_form = helpers.ActionForm
    actions_on_top = True
    actions_on_bottom = False
    actions_selection_counter = True
    
    #the following proxy to the proper schema admin
    create = views.DocumentProxyView
    update = views.DocumentProxyView
    
    delete = views.DeleteView
    index = views.IndexView
    history = views.HistoryView
    detail_views = [views.HistoryView, views.DeleteView]
    default_schema_admin = SchemaAdmin
    schema_inlines = [] # [(Schema, SchemaAdminCls),]
    
    def get_urls(self):
        def wrap(view, cacheable=False):
            def wrapper(*args, **kwargs):
                return self.as_view(view, cacheable)(*args, **kwargs)
            return update_wrapper(wrapper, view)
        
        init = self.get_view_kwargs()
        
        # Admin-site-wide views.
        urlpatterns = self.get_extra_urls()
        urlpatterns += patterns('',
            url(r'^$',
                wrap(self.index.as_view(**init)),
                name=self.app_name+'_changelist'),
            url(r'^add/$',
                wrap(self.create.as_view(**init)),
                name=self.app_name+'_add'),
        )
        
        for key, view in self.get_detail_views().iteritems():
            url_p = r'^(?P<pk>.+)/%s/$' % key
            if hasattr(view, 'get_url_pattern'):
                url_p = view.get_url_pattern(self)
            urlpatterns += patterns('',
                url(url_p,
                    wrap(view.as_view(**init)),
                    name='%s_%s' % (self.app_name, key)),
            )
        
        urlpatterns += patterns('', #we shouldn't put anything after this url
            url(r'^(?P<pk>.+)/$',
                wrap(self.update.as_view(**init)),
                name=self.app_name+'_change'),
        )
        return urlpatterns
    
    def urls(self):
        return self.get_urls(), self.app_name, None
    urls = property(urls)
    
    def get_extra_urls(self):
        return patterns('',)
    
    def get_detail_views(self):
        ret = dict()
        for detail_view in self.detail_views:
            ret[detail_view.key] = detail_view
        return ret
    
    def get_changelist(self, request):
        from changelist import ChangeList
        return ChangeList
    
    def reverse(self, name, *args, **kwargs):
        from django.core.urlresolvers import get_urlconf, get_resolver
        urlconf = get_urlconf()
        resolver = get_resolver(urlconf)
        app_list = resolver.app_dict['admin']
        return reverse('%s:%s' % (self.admin_site.name, name), args=args, kwargs=kwargs, current_app=self.app_name)
    
    def create_admin_for_schema(self, schema):
        admin_class = self.get_admin_class_for_schema(schema)
        return admin_class(self.model, self.admin_site, schema, self)
    
    def get_admin_class_for_schema(self, schema):
        for cls, admin_class in self.schema_inlines:
            if schema == cls:
                return admin_class
        return self.default_schema_admin
    
    def log_addition(self, request, object):
        """
        Log that an object has been successfully added.

        The default implementation creates an admin LogEntry object.
        """
        from django.contrib.admin.models import LogEntry, ADDITION
        LogEntry.objects.log_action(
            user_id         = request.user.pk,
            #content_type_id = ContentType.objects.get_for_model(object).pk,
            content_type_id = None,
            object_id       = object.get_id(),
            object_repr     = force_unicode(object),
            action_flag     = ADDITION
        )
    
    def log_change(self, request, object, message):
        """
        Log that an object has been successfully changed.

        The default implementation creates an admin LogEntry object.
        """
        from django.contrib.admin.models import LogEntry, CHANGE
        LogEntry.objects.log_action(
            user_id         = request.user.pk,
            #content_type_id = ContentType.objects.get_for_model(object).pk,
            content_type_id = None,
            object_id       = object.get_id(),
            object_repr     = force_unicode(object),
            action_flag     = CHANGE,
            change_message  = message
        )
    
    def log_deletion(self, request, object, object_repr):
        """
        Log that an object will be deleted. Note that this method is called
        before the deletion.

        The default implementation creates an admin LogEntry object.
        """
        from django.contrib.admin.models import LogEntry, DELETION
        LogEntry.objects.log_action(
            user_id         = request.user.pk,
            #content_type_id = ContentType.objects.get_for_model(self.model).pk,
            content_type_id = None,
            object_id       = object.get_id(),
            object_repr     = object_repr,
            action_flag     = DELETION
        )
    
    def get_base_breadcrumbs(self):
        admin_name = self.admin_site.name
        model_name = self.model._meta.verbose_name
        opts = self.model._meta
        breadcrumbs = [
            Breadcrumb('Home', ['%s:index' % admin_name]),
            Breadcrumb(opts.app_label, ['%s:app_list' % admin_name, (self.app_name,), {}]),
            Breadcrumb(opts.verbose_name_plural, ['%s:%s_changelist' % (admin_name, self.app_name)]),
        ]
        return breadcrumbs
    
    def get_instance_breadcrumb(self, obj=None):
        if obj:
            return Breadcrumb(unicode(obj))
        return Breadcrumb('Add %s' % self.model._meta.verbose_name)
    
    def has_add_permission(self, request):
        opts = self.model._meta
        #print opts.collection, request.user.has_perm('dockit.'+ opts.get_add_permission()), opts.get_add_permission()
        #print request.user.user_permissions.all().filter(codename=opts.get_add_permission()).values_list('codename', 'content_type')
        return request.user.has_perm('dockit.'+ opts.get_add_permission())
    
    def has_change_permission(self, request, obj=None):
        opts = self.model._meta
        return request.user.has_perm('dockit.'+ opts.get_change_permission(), obj)
    
    def has_delete_permission(self, request, obj=None):
        opts = self.model._meta
        return request.user.has_perm('dockit.'+ opts.get_delete_permission(), obj)




from django.core.urlresolvers import reverse

class Breadcrumb(object):
    def __init__(self, name, url=None):
        self.name = name
        self.url = url
    
    def get_absolute_url(self):
        if not isinstance(self.url, basestring):
            if len(self.url) > 1: #hack
                return reverse(self.url[0], args=self.url[1], kwargs=self.url[2])
            return reverse(*self.url)
        return self.url




from django.conf import settings

backend = None

DYNAMIC_IMPORT_CACHE = dict()

def dynamic_import(name):
    if not isinstance(name, str):
        return name
    if name in DYNAMIC_IMPORT_CACHE:
        return DYNAMIC_IMPORT_CACHE[name]
    original_name = name
    names = name.split('.')
    attr = names.pop()
    try:
        ret = __import__(name, globals(), locals(), [attr], -1)
    except ImportError:
        name = '.'.join(names)
        ret = __import__(name, globals(), locals(), [attr], -1)
    ret = getattr(ret, attr, ret)
    DYNAMIC_IMPORT_CACHE[original_name] = ret
    return ret

def get_document_backend():
    #TODO be thread safe
    global backend
    if not backend:
        backend_path = getattr(settings, 'DOCKIT_BACKEND', 'dockit.backends.djangodocument.backend.ModelDocumentStorage')
        backend = dynamic_import(backend_path)
        if isinstance(backend, type) or callable(backend):
            backend = backend()
    return backend




from django.db import models

class DocumentManager(models.Manager):
    def __init__(self, *args, **kwargs):
        super(DocumentManager, self).__init__(*args, **kwargs)
        self.index_models = dict()
    
    def register_index_model(self, key, model, instance_types):
        self.index_models[key] = {'model':model,
                                  'instance_types':instance_types}
    
    def get_index(self, key):
        return self.index_models[key]
    
    def filter_by_indexes(self, mapping, **filters):
        """
        mapping is a dictionary of the index as keys, and the datatype as the value
        """
        qs = self.all()
        for param, value in filters.iteritems():
            index = param.split('__', 1)[0]
            if index in mapping:
                key = mapping[index]
                qs &= self.filter_by_indexed_value(key, param, value)
            else:
                raise KeyError, "Unrecognized filter: %s" % index
        return qs
    
    def filter_by_indexed_value(self, key, index, value):
        model = self.get_index(key)['model']
        return self.filter(**model.objects.filter_kwargs_for_value(index, value))

class BaseIndexManager(models.Manager):
    def filter_kwargs_for_operation(self, operation):
        if operation.key in ('pk', '_pk'):
            return {'pk__%s' % operation.operation: operation.value}
        prefix = self.model._meta.get_field('document').related.var_name
        filter_kwargs = dict()
        filter_kwargs['%s__index' % prefix] = operation.key
        filter_kwargs['%s__value__%s' % (prefix, operation.operation)] = operation.value
        return filter_kwargs
    
    def unique_values(self, index):
        return self.filter(index=index).values_list('value', flat=True).distinct()
    
    def clear_db_index(self, instance_id, name=None):
        if name is None:
            return self.filter(document=instance_id).delete()
        return self.filter(document=instance_id, index=name).delete()
    
    def db_index(self, instance_id, name, value):
        self.filter(document=instance_id, index=name).delete()
        from dockit.schema import Document
        if isinstance(value, models.Model):
            value = value.pk
        if isinstance(value, Document):
            value = value.pk
        obj = self.create(document_id=instance_id, index=name, value=value)




from dockit import schema
from dockit.backends import get_document_backend

from django.db import models
from django.conf import settings

from copy import deepcopy

import datetime

class SchemaProxyDict(dict):
    def __getitem__(self, key):
        val = dict.__getitem__(self, key)
        return TemporaryDocument.generate_document(val)

class TemporaryDocument(schema.Document):
    @classmethod
    def generate_document(cls, document):
        class GeneratedTempDocument(cls):
            class Meta:
                proxy = True
        fields = deepcopy(document._meta.fields)
        
        #handle dynamic typing
        if document._meta.typed_field:
            GeneratedTempDocument._meta.typed_field = document._meta.typed_field
            GeneratedTempDocument._meta.typed_key = document._meta.typed_key
            t_field = fields[document._meta.typed_field]
            orignal_schemas = t_field.schemas
            t_field.schemas = SchemaProxyDict(orignal_schemas)
        
        GeneratedTempDocument._meta.fields.update(fields)
        GeneratedTempDocument._meta.original_document = document
        
        for name, value in GeneratedTempDocument._meta.fields.iteritems():
            if hasattr(value, 'contribute_to_class'):
                value.contribute_to_class(GeneratedTempDocument, name)
            else:
                setattr(GeneratedTempDocument, name, value)
        
        return GeneratedTempDocument
    
    def commit_changes(self, doc_id=None):
        document_cls = self._meta.original_document
        
        backend = get_document_backend()
        id_field = backend.get_id_field_name()
        
        data = self.to_primitive(self)
        data[id_field] = doc_id
        
        instance = document_cls(_primitive_data=data)
        instance.save()
        return instance
    
    @classmethod
    def create_from_instance(cls, instance):
        backend = get_document_backend()
        data = instance.to_primitive(instance)
        instance_id = data.pop(backend.get_id_field_name(), None)
        obj = cls.to_python(data)
        obj._original_id = instance_id
        return obj

def create_temporary_document_class(document_cls):
    return TemporaryDocument.generate_document(document_cls)

#TODO creating an index should optionally populate this table
class ActiveIndex(schema.Document):
    collection = schema.CharField()
    name = schema.CharField() #name assigned in application
    backend = schema.CharField()
    backend_index_identifier = schema.CharField()
    creation_date = schema.DateTimeField(default=datetime.datetime.now)
    #TODO store the index parameters

class DockitPermission(models.Model):
    pass #a content type is required for creating permissions

if 'django.contrib.auth' in settings.INSTALLED_APPS:
    import auth




from django.db import models

import datetime
from decimal import Decimal

from managers import BaseIndexManager, DocumentManager

class DocumentStore(models.Model):
    collection = models.CharField(max_length=128)
    data = models.TextField()
    
    objects = DocumentManager()
    
    def clear_indexes(self):
        for index in type(self).objects.index_models.itervalues():
            index['model'].objects.clear_db_index(self)

class BaseIndex(models.Model):
    document = models.ForeignKey(DocumentStore)
    index = models.CharField(max_length=50, db_index=True)
    timestamp = models.DateTimeField(auto_now=True)
    
    objects = BaseIndexManager()
    
    class Meta:
        abstract = True

class IntegerIndex(BaseIndex):
    value = models.IntegerField(null=True)
DocumentStore.objects.register_index_model('int', IntegerIndex, int)

class BooleanIndex(BaseIndex):
    value = models.BooleanField()
DocumentStore.objects.register_index_model('bool', BooleanIndex, bool)

class StringIndex(BaseIndex):
    value = models.CharField(max_length=512)
DocumentStore.objects.register_index_model('char', StringIndex, basestring)

class TextIndex(BaseIndex):
    value = models.TextField()
DocumentStore.objects.register_index_model('text', TextIndex, basestring)

class DateTimeIndex(BaseIndex):
    value = models.DateTimeField(null=True)
DocumentStore.objects.register_index_model('datetime', DateTimeIndex, datetime.datetime)

class DateIndex(BaseIndex):
    value = models.DateField(null=True)
DocumentStore.objects.register_index_model('date', DateIndex, datetime.date)

class FloatIndex(BaseIndex):
    value = models.FloatField(null=True)
DocumentStore.objects.register_index_model('float', FloatIndex, float)

class TimeIndex(BaseIndex):
    value = models.TimeField(null=True)
DocumentStore.objects.register_index_model('time', TimeIndex, datetime.time)

class DecimalIndex(BaseIndex):
    value = models.DecimalField(max_digits=19, decimal_places=10, null=True)
DocumentStore.objects.register_index_model('decimal', DecimalIndex, Decimal)





class AdminViewMixin:
    admin = None
    admin_site = None
    
    def is_popup(self):
        return "_popup" in self.request.REQUEST
    
    def get_context_data(self, **kwargs):
        return {'media': self.admin.media,
                'admin': self.admin,
                'admin_site': self.admin_site,
                'is_popup': self.is_popup()}





from django.utils.encoding import smart_unicode, force_unicode
from django.utils.text import capfirst
from django.utils import formats
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.fields import BLANK_CHOICE_DASH
from django import forms

from decimal import Decimal
import datetime
import copy

from serializer import PRIMITIVE_PROCESSOR
from exceptions import DotPathNotFound
from common import DotPathList, DotPathDict, DotPathSet, UnSet
from dockit.forms.fields import HiddenSchemaField, HiddenListField, HiddenDictField, SchemaChoiceField, PrimitiveListField, SchemaMultipleChoiceField

class NOT_PROVIDED:
    pass

class BaseField(object):
    form_field_class = forms.CharField
    form_field_choices_class = forms.ChoiceField
    form_widget_class = None
    
    creation_counter = 0
    auto_creation_counter = -1
    
    def __init__(self, verbose_name=None, name=None, blank=False, null=False,
                 default=NOT_PROVIDED, editable=True,
                 serialize=True, choices=None, help_text='',
                 validators=[],
                 db_index=False, unique=False,
                #db_index=False,  db_column=None, primary_key=False, max_length=None
                #db_tablespace=None, auto_created=False, 
                #unique_for_date=None, unique_for_month=None, unique_for_year=None, 
                 error_messages=None, auto_created=False):
        self.verbose_name = verbose_name
        self.name = name
        self.blank = blank
        self.null = null
        self.default = default
        self.editable = editable
        self.serialize = serialize,
        self.choices = choices
        self.help_text = help_text
        self.validators = validators
        self.db_index = db_index
        
        #TODO support the following:
        self.rel = None
        self.flatchoices = None
        self.unique = unique
        
        # Adjust the appropriate creation counter, and save our local copy.
        if auto_created:
            self.creation_counter = BaseField.auto_creation_counter
            BaseField.auto_creation_counter -= 1
        else:
            self.creation_counter = BaseField.creation_counter
            BaseField.creation_counter += 1
    
    def contribute_to_class(self, cls, name):
        self.name = name
        if not self.verbose_name:
            self.verbose_name = name.replace('_', ' ')
        cls._meta.fields[name] = self
        setattr(cls, name, self)
    
    def has_default(self):
        "Returns a boolean of whether this field has a default value."
        return self.default is not NOT_PROVIDED

    def get_default(self):
        "Returns the default value for this field."
        if self.has_default():
            if callable(self.default):
                return self.default()
            return force_unicode(self.default, strings_only=True)
        if not self.blank or self.null:
            return None
        return ""
    
    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH):
        """Returns choices with a default blank choices included, for use
        as SelectField choices for this field."""
        first_choice = include_blank and blank_choice or []
        if self.choices:
            choices = self.choices
            if callable(choices):
                choices = choices()
            return first_choice + list(choices)
        return first_choice
        #rel_model = self.rel.to
        #if hasattr(self.rel, 'get_related_field'):
        #    lst = [(getattr(x, self.rel.get_related_field().attname), smart_unicode(x)) for x in rel_model._default_manager.complex_filter(self.rel.limit_choices_to)]
        #else:
        #lst = [(x._get_pk_val(), smart_unicode(x)) for x in rel_model._default_manager.complex_filter(self.rel.limit_choices_to)]
        #return first_choice + lst

    def get_choices_default(self):
        return self.get_choices()
    
    def to_primitive(self, val):
        """
        Takes a python object and returns an object that is [json] serializable
        """
        return val
    
    def to_python(self, val, parent=None):
        return val
    
    def get_form_field_class(self):
        if self.choices:
            return self.form_field_choices_class
        else:
            return self.form_field_class
    
    def formfield(self, **kwargs):
        "Returns a django.forms.Field instance for this database Field."
        default = self.get_form_field_class()
        form_class = kwargs.pop('form_class', default)
        defaults = self.formfield_kwargs(**kwargs)
        return form_class(**defaults)
    
    def formfield_kwargs(self, **kwargs):
        defaults = {'required': not self.blank,
                    'label': capfirst(self.verbose_name),
                    'help_text': self.help_text,
                    'widget':self.form_widget_class,}
        if self.has_default():
            defaults['initial'] = self.get_default()
        
        for key, value in defaults.items():
            if value is None:
                del defaults[key]
        
        if self.choices:
            # Fields with choices get special treatment.
            include_blank = self.blank or not (self.has_default() or 'initial' in kwargs)
            defaults['choices'] = self.get_choices(include_blank=include_blank)
            #defaults['coerce'] = self.to_python
            #if self.null:
            #    defaults['empty_value'] = None
            # Many of the subclass-specific formfield arguments (min_value,
            # max_value) don't apply for choice fields, so be sure to only pass
            # the values that TypedChoiceField will understand.
            for k in kwargs.keys():
                if k not in ('empty_value', 'choices', 'required',#'coerce',
                             'widget', 'label', 'initial', 'help_text',
                             'error_messages', 'show_hidden_initial'):
                    del kwargs[k]
        defaults.update(kwargs)
        return defaults
    
    def traverse_dot_path(self, traverser):
        traverser.end(field=self)
    
    def is_instance(self, value):
        return False
    
    def set_value(self, parent, attr, value):
        raise DotPathNotFound
    
    def __deepcopy__(self, memodict):
        obj = copy.copy(self)
        memodict[id(self)] = obj
        return obj
    
    def _get_val_from_obj(self, obj):
        return obj[self.name]
    
    def value_to_string(self, obj):
        val = self._get_val_from_obj(obj)
        return self.to_primitive(val)
    
    @property
    def attname(self):
        return self.name

class BaseTypedField(BaseField):
    coerce_function = None
    
    def to_primitive(self, val):
        if isinstance(self.coerce_function, type) and isinstance(val, self.coerce_function):
            return val
        if val is None:
            return val
        return self.coerce_function(val)
    
    def is_instance(self, val):
        if isinstance(self.coerce_function, type) and isinstance(val, self.coerce_function):
            return True
        if val is None:
            return True
        return False

class CharField(BaseTypedField):
    coerce_function = unicode

class TextField(BaseTypedField):
    coerce_function = unicode
    form_widget_class = forms.Textarea

class IntegerField(BaseTypedField):
    coerce_function = int
    form_field_class = forms.IntegerField

class BigIntegerField(BaseTypedField):
    coerce_function = long
    form_field_class = forms.IntegerField

class BooleanField(BaseTypedField):
    coerce_function = bool
    form_field_class = forms.BooleanField
    
    def __init__(self, *args, **kwargs):
        super(BooleanField, self).__init__(*args, **kwargs)
        self.blank = True

class DateField(BaseField):
    form_field_class = forms.DateField
    
    def __init__(self, *args, **kwargs):
        self.input_formats = kwargs.pop('input_formats', list())
        super(DateField, self).__init__(*args, **kwargs)
    
    def to_python(self, value, parent=None):
        """
        Validates that the input can be converted to a date. Returns a Python
        datetime.date object.
        """
        if value is None:
            return None
        if isinstance(value, datetime.datetime):
            return value.date()
        if isinstance(value, datetime.date):
            return value
        for format in self.input_formats or formats.get_format('DATE_INPUT_FORMATS'):
            try:
                return datetime.datetime.strptime(value, format).date()
            except ValueError:
                continue
        raise ValueError

class DateTimeField(BaseField):
    form_field_class = forms.DateTimeField
    
    def __init__(self, *args, **kwargs):
        self.input_formats = kwargs.pop('input_formats', list())
        super(DateTimeField, self).__init__(*args, **kwargs)
    
    def to_python(self, value, parent=None):
        """
        Validates that the input can be converted to a date. Returns a Python
        datetime.date object.
        """
        if value is None:
            return None
        if isinstance(value, datetime.datetime):
            return value
        if isinstance(value, datetime.date):
            return value
        for format in self.input_formats or formats.get_format('DATE_INPUT_FORMATS'):
            try:
                return datetime.datetime.strptime(value, format)
            except ValueError:
                continue
        raise ValueError

class DecimalField(BaseField):
    form_field_class = forms.DecimalField
    
    def __init__(self, *args, **kwargs):
        self.max_digits = kwargs.pop('max_digits', None)
        self.decimal_places = kwargs.pop('decimal_places', None)
        super(DecimalField, self).__init__(*args, **kwargs)
    
    def to_primitive(self, val):
        return str(val)
    
    def to_python(self, val, parent=None):
        return Decimal(val)

class EmailField(CharField):
    form_field_class = forms.EmailField
    pass #TODO validate

#TODO filefield? this will be passed off to the backend

class FloatField(BaseTypedField):
    coerce_function = float
    form_field_class = forms.FloatField

#TODO imagefield

class IPAddressField(CharField):
    form_field_class = forms.IPAddressField
    #TODO validate

#class GenericIPAddressField(TextField):
#    form_field_class = forms.GenericIPAddressField
#    #TODO validate

#TODO NullBooleanField

class PositiveIntegerField(IntegerField):
    form_field_class = forms.IntegerField
    #TODO validate

#TODO PositiveSmallIntegerField

class SlugField(CharField):
    form_field_class = forms.SlugField

#TODO SmallIntegerField

class TimeField(BaseTypedField):
    coerce_function = datetime.time
    form_field_class = forms.TimeField

#TODO URLField
#TODO XMLField

class SchemaTypeField(CharField):
    form_widget_class = forms.HiddenInput
    
    def __init__(self, schemas, *args, **kwargs):
        self.schemas = schemas
        super(SchemaTypeField, self).__init__(*args, **kwargs)
    
    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH):
        """Returns choices with a default blank choices included, for use
        as SelectField choices for this field."""
        first_choice = include_blank and blank_choice or []
        if self.schemas:
            return first_choice + list(zip(self.schemas.keys(), self.schemas.keys()))
            #TODO allow schema to have a display label, perhaps verbose name?
        return first_choice

class SchemaField(BaseField):
    form_field_class = HiddenSchemaField
    
    def __init__(self, schema, *args, **kwargs):
        self.schema = schema
        super(SchemaField, self).__init__(*args, **kwargs)
    
    def to_primitive(self, val):
        return self.schema.to_primitive(val)
    
    def to_python(self, val, parent=None):
        return self.schema.to_python(val, parent)
    
    def is_instance(self, val):
        if val is None:
            return True
        return isinstance(val, self.schema)
    
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            name = traverser.next_part
            value = traverser.current_value
            next_value = field = None
            schema = self.schema
            
            if value:
                try:
                    next_value = value[name]
                except KeyError:
                    pass
                schema = type(value)
            
            if name in schema._meta.fields:
                field = schema._meta.fields[name]
            
            traverser.next(field=field, value=next_value)
        else:
            traverser.end(field=self)
    
    def set_value(self, parent, attr, value):
        if value is UnSet:
            del parent[attr]
        else:
            parent[attr] = value

class GenericSchemaField(BaseField): #meant to be sublcassed
    def __init__(self, field_name='_type', **kwargs):
        self.type_field_name = field_name
        super(GenericSchemaField, self).__init__(**kwargs)
    
    def lookup_schema(self, key):
        raise NotImplementedError
    
    def get_schema_type(self, val):
        return val._meta.schema_key
    
    def get_schema_choices(self):
        raise NotImplementedError
    
    def set_schema_type(self, val):
        val[self.type_field_name] = self.get_schema_type(val)
    
    def to_primitive(self, val):
        if hasattr(val, 'to_primitive'):
            self.set_schema_type(val)
            return val.to_primitive(val)
        return val
    
    def to_python(self, val, parent=None):
        if hasattr(val, 'to_python'):
            return val.to_python(val, parent)
        if not self.is_instance(val):
            key = val[self.type_field_name]
            schema_cls = self.lookup_schema(key)
            return schema_cls(_primitive_data=val)
        return val
    
    def is_instance(self, val):
        if val is None:
            return True
        from schema import Schema
        return isinstance(val, Schema)
    
    def traverse_dot_path(self, traverser):
        #CONSIDER if there is a value then return an exacto schema field
        if traverser.remaining_paths:
            name = traverser.next_part
            value = traverser.current_value
            next_value = field = None
            
            if value:
                try:
                    next_value = value[name]
                except KeyError:
                    pass
                
                if hasattr(value, '_meta') and hasattr(value._meta, 'fields') and name in value._meta.fields:
                    field = value._meta.fields[name]
            
            traverser.next(field=field, value=next_value)
        else:
            if traverser.current_value:
                traverser.end(field=SchemaField(schema=type(traverser.current_value)))
            else:
                traverser.end(field=self)
    
    def set_value(self, parent, attr, value):
        if value is UnSet:
            del parent[attr]
        else:
            parent[attr] = value

class TypedSchemaField(GenericSchemaField):
    def __init__(self, schemas, field_name='_type', **kwargs):
        self.schemas = schemas
        super(TypedSchemaField, self).__init__(field_name, **kwargs)
    
    def lookup_schema(self, key):
        return self.schemas[key]
    
    def get_schema_type(self, val):
        return val._meta.schema_key
    
    def get_schema_choices(self):
        return self.schemas.items()
    
    #TODO what about a widget?

#TODO need a more comprehensive form field and widget solution for these
class ListField(BaseField):
    form_field_class = HiddenListField
    
    def __init__(self, subfield=None, *args, **kwargs):
        self.subfield = subfield
        kwargs.setdefault('default', list)
        super(ListField, self).__init__(*args, **kwargs)
    
    def get_form_field_class(self):
        if not hasattr(self.subfield, 'schema'):
            return PrimitiveListField
        else:
            return self.form_field_class
    
    def formfield_kwargs(self, **kwargs):
        kwargs = super(ListField, self).formfield_kwargs(**kwargs)
        if not hasattr(self.subfield, 'schema'):
            kwargs['subfield'] = self.subfield.formfield()
        return kwargs
    
    def to_primitive(self, val):
        if self.subfield:
            ret = list()
            if val is None:
                return ret
            for item in val:
                ret.append(self.subfield.to_primitive(item))
            #run data through the primitive processor
            return PRIMITIVE_PROCESSOR.to_primitive(ret)
        return PRIMITIVE_PROCESSOR.to_primitive(val)
    
    def to_python(self, val, parent=None):
        if self.subfield:
            ret = DotPathList()
            if val is None:
                return ret
            #TODO pass in parent
            for item in val:
                if not self.subfield.is_instance(item):
                    item = self.subfield.to_python(item)
                ret.append(item)
            #run data through the primitive processor
            return PRIMITIVE_PROCESSOR.to_python(ret)
        return PRIMITIVE_PROCESSOR.to_python(val)
    
    def is_instance(self, val):
        if val is None:
            return True
        if not isinstance(val, DotPathList):
            return False
        if self.subfield:
            for item in val:
                if not self.subfield.is_instance(item):
                    return False
        return True
    
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            new_value = None
            value = traverser.current_value
            name = traverser.next_part
            if value and name != '*':
                try:
                    new_value = value[int(name)]
                except ValueError:
                    raise DotPathNotFound(u"Invalid index given, must be an integer (%s)" % name)
                except IndexError:
                    pass
            elif value and name == '*':
                new_value = value #requested a list of values
            traverser.next(field=self.subfield, value=new_value)
        else:
            traverser.end(field=self)
    
    def set_value(self, parent, attr, value):
        index = int(attr)
        if value is UnSet:
            parent.pop(index)
        else:
            if self.subfield and not self.subfield.is_instance(value):
                value = self.subfield.to_python(value)
            if index == len(parent):
                parent.append(value)
            else:
                parent[index] = value

class SetField(ListField):
    form_field_choices_class = forms.MultipleChoiceField
    
    #returns a MultipleChoiceField if there is a set of choices
    def get_form_field_class(self):
        if self.choices:
            return self.form_field_choices_class
        return ListField.get_form_field_class(self)
    
    def formfield_kwargs(self, **kwargs):
        if self.choices:
            return BaseField.formfield_kwargs(self, **kwargs)
        else:
            return ListField.formfield_kwargs(self, **kwargs)
    
    def to_python(self, val, parent=None):
        if self.subfield:
            ret = DotPathSet()
            if val is None:
                return ret
            #TODO pass in parent
            for item in val:
                if not self.subfield.is_instance(item):
                    item = self.subfield.to_python(item)
                ret.add(item)
            #run data through the primitive processor
            return PRIMITIVE_PROCESSOR.to_python(ret)
        return PRIMITIVE_PROCESSOR.to_python(val)
    
    def is_instance(self, val):
        if val is None:
            return True
        if not isinstance(val, DotPathSet):
            return False
        if self.subfield:
            for item in val:
                if not self.subfield.is_instance(item):
                    return False
        return True

class DictField(BaseField):
    form_field_class = HiddenDictField
    
    def __init__(self, key_subfield=None, value_subfield=None, **kwargs):
        self.key_subfield = key_subfield
        self.value_subfield = value_subfield
        super(DictField, self).__init__(**kwargs)
    
    def to_primitive(self, val):
        ret = dict()
        if val is None:
            return ret
        for key, value in val.iteritems():
            if hasattr(value, 'to_primitive'):
                value = type(value).to_primitive(value)
            if hasattr(key, 'to_primitive'):
                key = type(key).to_primitive(key)
            ret[key] = value
        #TODO run data through the primitive processor
        ret = PRIMITIVE_PROCESSOR.to_primitive(ret)
        return ret
    
    def to_python(self, val, parent=None):
        ret = DotPathDict()
        if val is None:
            return ret
        for key, value in val.iteritems():
            if self.value_subfield:
                if not self.value_subfield.is_instance(value):
                    value = self.value_subfield.to_python(value)
            if self.key_subfield:
                if not self.key_subfield.is_instance(key):
                    key = self.key_subfield.to_python(key)
            ret[key] = value
        #TODO run data through the primitive processor
        ret = PRIMITIVE_PROCESSOR.to_python(ret)
        return ret
    
    def is_instance(self, val):
        if val is None:
            return True
        if not isinstance(val, DotPathDict):
            return False
        if self.value_subfield:
            for item in val.itervalues():
                if not self.value_subfield.is_instance(item):
                    return False
        if self.key_subfield:
            for item in val.iterkeys():
                if not self.key_subfield.is_instance(item):
                    return False
        return True
    
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            value = traverser.current_value
            name = traverser.next_part
            new_value = None
            if value and name != '*':
                try:
                    new_value = value[name]
                except KeyError:
                    pass
                if new_value and not hasattr(new_value, 'traverse_dot_path'):
                    from common import DotPathDict
                    new_value = DotPathDict(new_value)
            traverser.next(field=self.value_subfield, value=new_value)
        else:
            traverser.end(field=self)
    
    def set_value(self, parent, attr, value):
        if value is UnSet:
            del parent[attr]
        else:
            if self.value_subfield and not self.value_subfield.is_instance(value):
                value = self.value_subfield.to_python(value)
            parent[attr] = value

class ReferenceField(BaseField):
    form_field_class = SchemaChoiceField
    
    def __init__(self, document, *args, **kwargs):
        if document == 'self':
            self.self_reference = True
            document = None
        else:
            self.self_reference = False
            assert hasattr(document, 'objects')
            assert hasattr(document, 'get_id')
        self.document = document
        super(ReferenceField, self).__init__(*args, **kwargs)
    
    @property
    def _meta(self):
        return self.document._meta
    
    def is_instance(self, val):
        if val is None:
            return True
        return isinstance(val, self.document)
    
    def to_primitive(self, val):
        if isinstance(val, basestring): #CONSIDER, should this happen?
            return val
        if val is None:
            return val
        return val.get_id()
    
    def to_python(self, val, parent=None):
        if self.self_reference:
            if val is None:
                return val
            document = type(parent)
            if isinstance(val, document):
                return val
        else:
            if self.is_instance(val):
                return val
            document = self.document
        try:
            return document.objects.get(pk=val)
        except ObjectDoesNotExist:
            if self.null:
                return None
            raise
    
    def formfield_kwargs(self, **kwargs):
        kwargs = BaseField.formfield_kwargs(self, **kwargs)
        kwargs.setdefault('queryset', self.document.objects.all())
        return kwargs
    
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            name = traverser.next_part
            value = traverser.current_value
            next_value = field = None
            
            if value:
                try:
                    next_value = value[name]
                except KeyError:
                    pass
            if name in self.document._meta.fields:
                field = self.document._meta.fields[name]
            
            traverser.next(field=field, value=next_value)
        else:
            traverser.end(field=self)
    
    def set_value(self, parent, attr, value):
        parent[attr] = value

class DocumentSetField(SetField):
    form_field_choices_class = SchemaMultipleChoiceField
    
    def __init__(self, document, *args, **kwargs):
        subfield = ReferenceField(document)
        kwargs['subfield'] = subfield
        super(DocumentSetField, self).__init__(*args, **kwargs)
        #self.choices = True
        self.document = document
        self.queryset = document.objects.all()
    
    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH):
        choices = list(include_blank and blank_choice or [])
        lst = [(x.pk, smart_unicode(x)) for x in self.queryset]
        return choices + lst
    
    def formfield_kwargs(self, **kwargs):
        kwargs = BaseField.formfield_kwargs(self, **kwargs)
        kwargs.setdefault('queryset', self.queryset)
        return kwargs
    
    def get_form_field_class(self):
        return self.form_field_choices_class

class ModelReferenceField(BaseField):
    form_field_class = forms.ModelChoiceField
    form_field_choices_class = forms.ModelChoiceField
    
    def __init__(self, model, *args, **kwargs):
        self.model = model
        super(ModelReferenceField, self).__init__(*args, **kwargs)
    
    def is_instance(self, val):
        if val is None:
            return True
        return isinstance(val, self.model)
    
    def to_primitive(self, val):
        if val is None:
            return None
        return val.pk
    
    def to_python(self, val, parent=None):
        if val is None:
            return None
        return self.model.objects.get(pk=val)
    
    def formfield_kwargs(self, **kwargs):
        kwargs = BaseField.formfield_kwargs(self, **kwargs)
        kwargs.setdefault('queryset', self.model.objects)
        return kwargs

class ModelSetField(SetField):
    form_field_choices_class = forms.ModelMultipleChoiceField
    
    def __init__(self, model, *args, **kwargs):
        subfield = ModelReferenceField(model)
        kwargs['subfield'] = subfield
        super(ModelSetField, self).__init__(*args, **kwargs)
        #self.choices = True
        self.model = model
        self.queryset = model.objects.all()
    
    def get_form_field_class(self):
        return self.form_field_choices_class
    
    def formfield_kwargs(self, **kwargs):
        kwargs = BaseField.formfield_kwargs(self, **kwargs)
        kwargs.setdefault('queryset', self.model.objects)
        return kwargs
    
    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH):
        choices = list(include_blank and blank_choice or [])
        lst = [(x._get_pk_val(), smart_unicode(x)) for x in self.queryset]
        return choices + lst




from dockit.backends import get_document_backend

import re

from django.conf import settings
from django.db.models.options import get_verbose_name
from django.utils.translation import activate, deactivate_all, get_language, string_concat
from django.utils.encoding import smart_str, force_unicode
from django.utils.datastructures import SortedDict, MergeDict
from django.db.models import FieldDoesNotExist

from common import DotPathTraverser

class FieldsDict(MergeDict):
    def __init__(self, *dicts):
        self.fields = SortedDict()
        dicts = [self.fields] + list(dicts)
        super(FieldsDict, self).__init__(*dicts)
    
    def __setitem__(self, key, value):
        self.fields[key] = value
    
    def update(self, *args, **kwargs):
        return self.fields.update(*args, **kwargs)

class SchemaOptions(object):
    """ class based on django.db.models.options. We only keep
    useful bits."""
    
    abstract = True
    ordering = ['_id']
    
    DEFAULT_NAMES = ['verbose_name', 'db_table', 'ordering', 'schema_key',
                     'app_label', 'collection', 'virtual', 'proxy', 'permissions',
                     'typed_field', 'typed_key']
    
    def __init__(self, meta, app_label=None, parent_fields=[]):
        self.module_name, self.verbose_name = None, None
        self.verbose_name_plural = None
        self.object_name, self.app_label = None, app_label
        self.meta = meta
        self.fields = FieldsDict(*parent_fields)
        self.collection = None
        self.schema_key = None
        self.virtual = False #TODO all schemas are virtual
        self.proxy = False
        self._document = None
        self.typed_field = None
        self.typed_key = None
        self.permissions = []
    
    def process_values(self, cls):
        cls._meta = self
        self._document = cls
        self.installed = re.sub('\.models$', '', cls.__module__) in settings.INSTALLED_APPS
        # First, construct the default values for these options.
        self.object_name = cls.__name__
        self.module_name = self.object_name.lower()
        self.verbose_name = get_verbose_name(self.object_name)
        self.collection = self.default_schema_key()
        self.schema_key = self.default_schema_key()

        # Next, apply any overridden values from 'class Meta'.
        if getattr(self, 'meta', None):
            meta_attrs = self.meta.__dict__.copy()
            for name in self.meta.__dict__:
                # Ignore any private attributes that Django doesn't care about.
                # NOTE: We can't modify a dictionary's contents while looping
                # over it, so we loop over the *original* dictionary instead.
                if name.startswith('_'):
                    del meta_attrs[name]
            for attr_name in self.DEFAULT_NAMES + ['module_name']:
                if attr_name in meta_attrs:
                    setattr(self, attr_name, meta_attrs.pop(attr_name))
                elif hasattr(self.meta, attr_name):
                    setattr(self, attr_name, getattr(self.meta, attr_name))

            # verbose_name_plural is a special case because it uses a 's'
            # by default.
            setattr(self, 'verbose_name_plural', meta_attrs.pop('verbose_name_plural', string_concat(self.verbose_name, 's')))

            # Any leftover attributes must be invalid.
            if meta_attrs != {}:
                raise TypeError("'class Meta' got invalid attribute(s): %s" % ','.join(meta_attrs.keys()))
            del self.meta
        else:
            self.verbose_name_plural = string_concat(self.verbose_name, 's')
    
    def contribute_to_class(self, cls, name):
        #the following needs to happen after fields are populated
        if self.typed_field:
            if self.typed_key:
                if self.virtual:
                    raise TypeError("Virtual Schemas may not have a typed_key")
                #CONSIDER this will break if fields are registered after the meta, gennerally have the virtual create the field for you
                if self.typed_field not in self.fields:
                    raise TypeError("Non-virtual Schemas that specify a typed field and a typed key must have that typed field defined.")
                self.fields[self.typed_field].schemas[self.typed_key] = cls
                self.proxy = True
            else:
                #if not self.virtual:
                #    raise TypeError("Schemas that specify a typed_field and not a typed_key must be virtual.")
                if self.typed_field not in self.fields:
                    from fields import SchemaTypeField
                    self.polymorphic_schemas = SortedDict()
                    field = SchemaTypeField(self.polymorphic_schemas, editable=False)
                    field.contribute_to_class(cls, self.typed_field)
    
    def default_schema_key(self):
        return "%s.%s" % (smart_str(self.app_label), smart_str(self.module_name))
    
    def __str__(self):
        return self.collection

    def verbose_name_raw(self):
        """
        There are a few places where the untranslated verbose name is needed
        (so that we get the same value regardless of currently active
        locale).
        """
        lang = get_language()
        deactivate_all()
        raw = force_unicode(self.verbose_name)
        activate(lang)
        return raw
    verbose_name_raw = property(verbose_name_raw)
    
    def get_field(self, name):
        if name not in self.fields:
            if name == 'pk':
                return self.pk
            raise FieldDoesNotExist
        return self.fields[name]
    
    def get_field_by_name(self, name):
        """
        Returns the (field_object, model, direct, m2m), where field_object is
        the Field instance for the given name, model is the model containing
        this field (None for local fields), direct is True if the field exists
        on this model, and m2m is True for many-to-many relations. When
        'direct' is False, 'field_object' is the corresponding RelatedObject
        for this field (since the field doesn't have an instance associated
        with it).

        Uses a cache internally, so after the first access, this is very fast.
        """
        if name not in self.fields:
            raise FieldDoesNotExist
        return (self.fields[name], None, True, False)
    
    def get_ordered_objects(self):
        return []
    
    @property
    def pk(self):
        from fields import CharField
        return CharField(name='pk')
    
    def get_backend(self):
        return get_document_backend()
    
    def dot_notation_to_field(self, notation):
        traverser = DotPathTraverser(notation)
        traverser.resolve_for_schema(self._document)
        return traverser.current['field']
    
    def is_dynamic(self):
        return bool(self.typed_field)
    
    @property
    def local_fields(self):
        return self.fields.values()
    
    @property
    def many_to_many(self):
        return []
    
    def get_add_permission(self):
        return '%s.add' % self.collection

    def get_change_permission(self):
        return '%s.change' % self.collection

    def get_delete_permission(self):
        return '%s.delete' % self.collection

class DocumentOptions(SchemaOptions):
    abstract = False





import copy

from queryset import QuerySet



class QueryFilterOperation(object):
    def __init__(self, key, operation, value):
        self.key = key
        self.operation = operation
        from dockit.schema import Document
        from django.db.models import Model
        if isinstance(value, Model):
            value = value.pk
        if isinstance(value, Document):
            value = value.pk
        self.value = value
    
    def __hash__(self):
        return hash((self.key, self.operation, self.value))
    
    def dotpath(self):
        parts = self.key.split('__')
        return '.'.join(parts)

class QueryIndex(object):
    def __init__(self, document):
        self.name = None
        self.document = document
        self.inclusions = list()
        self.exclusions = list()
        self.indexes = list()
        
        self._queryset = None
    
    def _parse_key(self, key):
        #TODO detect and handle dotpath notation
        operation = 'exact'
        if '__' in key:
            key, operation = key.rsplit('__', 1)
        return key, operation
    
    def _parse_kwargs(self, kwargs):
        items = list()
        for key, value in kwargs.iteritems():
            #TODO proper parsing
            key, operation = self._parse_key(key)
            items.append(QueryFilterOperation(key=key, operation=operation, value=value))
        return items
    
    def _add_filter_parts(self, inclusions=[], exclusions=[], indexes=[]):
        new_index = type(self)(self.document)
        new_index.inclusions = self.inclusions + inclusions
        new_index.exclusions = self.exclusions + exclusions
        new_index.indexes = self.indexes + indexes
        return new_index
    
    def _build_queryset(self):
        backend = self.document._meta.get_backend()
        query = backend.get_query(self)
        return QuerySet(query)
    
    @property
    def queryset(self):
        if self._queryset is None:
            self._queryset = self._build_queryset()
        return self._queryset
    
    def filter(self, **kwargs):
        inclusions = self._parse_kwargs(kwargs)
        return self._add_filter_parts(inclusions=inclusions)
    
    def exclude(self, **kwargs):
        exclusions = self._parse_kwargs(kwargs)
        return self._add_filter_parts(exclusions=exclusions)
    
    def index(self, *args):
        items = list()
        for arg in args:
            key, operation = self._parse_key(arg)
            items.append(QueryFilterOperation(key=key, operation=operation, value=None))
        return self._add_filter_parts(indexes=items)
    
    def commit(self):
        backend = self.document._meta.get_backend()
        backend.register_index(self)
    
    def setname(self, name):
        self.name = name
    
    #proxy queryset methods
    def __len__(self):
        return self.queryset.__len__()
    
    def count(self):
        return self.__len__()
    
    def delete(self):
        return self.queryset.delete()
    
    def all(self):
        ret = copy.copy(self)
        ret._queryset = None
        return ret
    
    def get(self, **kwargs):
        return self.queryset.get(**kwargs)
    
    def exists(self):
        return self.queryset.exists()
    
    def __getitem__(self, val):
        return self.queryset.__getitem__(val)
    
    def __nonzero__(self):
        return self.queryset.__nonzero__()
    
    def __values__(self):
        return self.queryset.values()
    
    def __iter__(self):
        return self.queryset.__iter__()




from django.forms.widgets import Widget, Media, HiddenInput
from django.utils.safestring import mark_safe
import django.utils.copycompat as copy

from django import forms
from django.forms.util import flatatt
from django.forms.formsets import formset_factory, ORDERING_FIELD_NAME, DELETION_FIELD_NAME

class PrimitiveListWidget(Widget):
    '''
    Wraps around a subfield
    The widget receives the subfield and is responsible for rendering multiple iterations of the subfield and collecting of data
    '''
    def __init__(self, subfield, attrs=None):
        self.subfield = subfield
        super(PrimitiveListWidget, self).__init__(attrs)
    
    def get_base_form_class(self):
        class BaseForm(forms.Form):
            value = self.subfield
        return BaseForm
    
    def get_formset_class(self, **kwargs):
        form_cls = self.get_base_form_class()
        kwargs.setdefault('can_order', True)
        kwargs.setdefault('can_delete', True)
        formset = formset_factory(form_cls, **kwargs)
        return formset

    def render(self, name, value, attrs=None):
        if not isinstance(value, list):
            value = self.decompress(value)
        
        final_attrs = self.build_attrs(attrs)
        
        formset_class = self.get_formset_class()
        initial=[{'value':val} for val in value]
        formset = formset_class(initial=initial, prefix=name)
        parts = ['<div class="list-row form-row"><table>%s</table></div>' % form.as_table() for form in formset]
        parts.append('<div id="%s-empty" class="list-row form-row form-empty"><table>%s</table></div>' % (name, formset.empty_form.as_table()))
        output = u'<div%s style="float: left;" class="primitivelistfield" data-prefix="%s">%s %s</div>' % (flatatt(final_attrs), name, formset.management_form, u''.join(parts))
        return mark_safe(output)
    
    def value_from_datadict(self, data, files, name):
        formset_class = self.get_formset_class()
        formset = formset_class(data=data, files=files, prefix=name)
        value = list()
        for form in formset.forms:
            val = dict()
            for key in ('value', ORDERING_FIELD_NAME, DELETION_FIELD_NAME):
                val[key] = form.fields[key].widget.value_from_datadict(data, files, form.add_prefix(key))
            value.append(val)
        return value
    
    def _has_changed(self, initial, data):
        if initial is None:
            initial = [u'' for x in range(0, len(data))]
        else:
            if not isinstance(initial, list):
                initial = self.decompress(initial)
        #for widget, initial, data in zip(self.widgets, initial, data):
        #    if widget._has_changed(initial, data):
        #        return True
        return True #CONSIDER where is my name?
        return False
    
    def decompress(self, value):
        """
        Returns a list of decompressed values for the given compressed value.
        The given value can be assumed to be valid, but not necessarily
        non-empty.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def _get_media(self):
        "Media for a multiwidget is the combination of all media of the subwidgets"
        media = Media()
        media += self.subfield.media
        definition = getattr(self, 'Media', None)
        if definition:
            media += Media(definition)
        return media
    media = property(_get_media)
    
    def __deepcopy__(self, memo):
        obj = super(PrimitiveListWidget, self).__deepcopy__(memo)
        obj.subfield = copy.deepcopy(self.subfield)
        return obj




from pymongo import Connection
from pymongo.objectid import ObjectId

from dockit.backends.base import BaseDocumentStorage
from dockit.backends.queryset import BaseDocumentQuery

from django.conf import settings

class DocumentQuery(BaseDocumentQuery):
    def _build_params(self, include_indexes=False):
        params =  dict()
        for op in self.query_index.inclusions:
            indexer = self._get_indexer_for_operation(self.document, op)
            #i think this is horribly wrong
            params.update(indexer.filter())
        for op in self.query_index.exclusions:
            indexer = self._get_indexer_for_operation(self.document, op)
            params.update(indexer.filter())
        if include_indexes:
            for op in self.query_index.indexes:
                indexer = self._get_indexer_for_operation(self.document, op)
                params.update(indexer.filter())
        return params
    
    @property
    def collection(self):
        return self.document._meta.get_backend().get_collection(self.document._meta.collection)
    
    @property
    def queryset(self):
        params = self._build_params()
        if params:
            try:
                return self.collection.find(params)
            except TypeError:
                #why is it pymongo wants tuples for some and dictionaries for others?
                return self.collection.find(params.items())
        return self.collection.find()
    
    def wrap(self, entry):
        entry['_id'] = unicode(entry['_id'])
        return self.document.to_python(entry)
    
    def delete(self):
        params = self._build_params()
        if params:
            return self.collection.remove(params)
        return self.collection.remove()
    
    def get_from_filter_operations(self, filter_operations):
        params = self._build_params()
        for op in filter_operations:
            indexer = self._get_indexer_for_operation(self.document, op)
            params.update(indexer.filter())
        try:
            ret = self.collection.find_one(params)
        except TypeError:
            #why is it pymongo wants tuples for some and dictionaries for others?
            ret = self.collection.find_one(params.items())
        if ret is None:
            raise self.document.DoesNotExist
        return self.wrap(ret)
    
    def values(self):
        params = self._build_params(include_indexes=True)
        raise NotImplementedError
    
    def __len__(self):
        return self.queryset.count()
    
    def __nonzero__(self):
        return bool(self.queryset)
    
    def __getitem__(self, val):
        if isinstance(val, slice):
            results = list()
            #TODO i don't think mongo supports passing a slice
            for entry in self.queryset[val]:
                results.append(self.wrap(entry))
            return results
        else:
            return self.wrap(self.queryset[val])

class MongoDocumentStorage(BaseDocumentStorage):
    name = "mongodb"
    _indexers = dict() #TODO this should be automatic

    def __init__(self, username=None, password=None, host=None, port=None, db=None):
        #TODO be proper about this
        self.connection = Connection(host or settings.MONGO_HOST, port or settings.MONGO_PORT)
        self.db = self.connection[db or settings.MONGO_DB]
        username = getattr(settings, 'MONGO_USER', None)
        password = getattr(settings, 'MONGO_PASSWORD', None)
        if username:
            self.db.authenticate(username, password)
    
    def get_collection(self, collection):
        return self.db[collection]
    
    def save(self, doc_class, collection, data):
        id_field = self.get_id_field_name()
        if data.get(id_field, False) is None:
            del data[id_field]
        elif id_field in data:
            data[id_field] = ObjectId(data[id_field])
        self.get_collection(collection).save(data, safe=True)
        data[id_field] = unicode(data[id_field])
    
    def get(self, doc_class, collection, doc_id):
        data = self.get_collection(collection).find_one({'_id':ObjectId(doc_id)})
        if data is None:
            raise doc_class.DoesNotExist
        id_field = self.get_id_field_name()
        data[id_field] = unicode(data[id_field])
        return data
    
    def delete(self, doc_class, collection, doc_id):
        return self.get_collection(collection).remove(ObjectId(doc_id), safe=True)
    
    def get_id_field_name(self):
        return '_id'
    
    def register_index(self, query_index):
        query = self.get_query(query_index)
        params = query._build_params(include_indexes=True)
        for key in params.keys(): #TODO this is a hack
            params[key] = 1
        if params:
            collection = query_index.document._meta.collection
            try:
                self.get_collection(collection).ensure_index(params, background=True)
            except TypeError:
                self.get_collection(collection).ensure_index(params.items(), background=True)
    
    def get_query(self, query_index):
        return DocumentQuery(query_index)




from django.forms.fields import ChoiceField, Field, EMPTY_VALUES
from django.forms.widgets import HiddenInput, SelectMultiple, MultipleHiddenInput
from django.forms import ValidationError
from django.utils.translation import ugettext as _
from django.utils.encoding import force_unicode, smart_unicode
from django.utils import simplejson as json

from widgets import PrimitiveListWidget

import inspect

class HiddenJSONField(Field):
    widget = HiddenInput
    
    def to_python(self, value):
        if value and isinstance(value, basestring):
            try:
                return json.loads(value)
            except ValueError:
                raise ValidationError("Invalid JSON")#TODO more descriptive error?
        return None

    def validate(self, value):
        return Field.validate(self, value)
    
    def prepare_value(self, value):
        if hasattr(value, 'to_primitive'):
            return json.dumps(type(value).to_primitive(value))
        return Field.prepare_value(self, value)

class HiddenSchemaField(HiddenJSONField):
    pass

class HiddenListField(HiddenJSONField):
    pass

class HiddenDictField(HiddenJSONField):
    pass

class PrimitiveListField(Field):
    '''
    Wraps around a subfield
    The widget receives the subfield and is responsible for rendering multiple iterations of the subfield and collecting of data
    '''
    widget = PrimitiveListWidget
    
    def __init__(self, subfield, *args, **kwargs):
        self.subfield = subfield
        widget = kwargs.get('widget', self.widget)
        if isinstance(widget, type):
            widget = widget(subfield)
        kwargs['widget'] = widget
        super(PrimitiveListField, self).__init__(*args, **kwargs)
    
    def prepare_value(self, value):
        if value is None:
            return
        ret = list()
        for val in value:
            ret.append(self.subfield.prepare_value(val))
        return ret
    
    def bound_data(self, data, initial):
        ret = list()
        for data_item, initial_item in zip(data, initial):
            ret.append(self.subfield.bound_data(data_item, initial_item))
        return ret
    
    def clean(self, data, initial=None):
        ret = list()
        for i, data_item in enumerate(data):
            if data_item.get('DELETE', False):
                continue
            if initial and len(initial) > i:
                initial_item = initial[i]
            else:
                initial_item = None
            val = self.subfield.bound_data(data_item['value'], initial_item)
            
            arg_spec = inspect.getargspec(self.subfield.clean)
            try:
                if len(arg_spec.args) > 2:
                    val = self.subfield.clean(val, initial_item)
                else:
                    val = self.subfield.clean(val)
            except ValidationError:
                if val:
                    raise
            else:
                if val:
                    ret.append((val, data_item.get('ORDER', None)))
        
        def compare_ordering_key(k):
            if k[1] is None:
                return (1, 0) # +infinity, larger than any number
            return (0, k[1])
        
        ret.sort(key=compare_ordering_key)
        return [item[0] for item in ret]

class SchemaChoiceIterator(object):
    def __init__(self, field):
        self.field = field
        self.queryset = field.queryset

    def __iter__(self):
        if self.field.empty_label is not None:
            yield (u"", self.field.empty_label)
        if self.field.cache_choices:
            if self.field.choice_cache is None:
                self.field.choice_cache = [
                    self.choice(obj) for obj in self.queryset.all()
                ]
            for choice in self.field.choice_cache:
                yield choice
        else:
            for obj in self.queryset.all():
                yield self.choice(obj)

    def __len__(self):
        return len(self.queryset)

    def choice(self, obj):
        return (self.field.prepare_value(obj), self.field.label_from_instance(obj))

class SchemaChoiceField(ChoiceField):
    """A ChoiceField whose choices are a model QuerySet."""
    # This class is a subclass of ChoiceField for purity, but it doesn't
    # actually use any of ChoiceField's implementation.
    default_error_messages = {
        'invalid_choice': _(u'Select a valid choice. That choice is not one of'
                            u' the available choices.'),
    }

    def __init__(self, queryset, empty_label=u"---------", cache_choices=False,
                 required=True, widget=None, label=None, initial=None,
                 help_text=None, to_field_name=None, *args, **kwargs):
        if required and (initial is not None):
            self.empty_label = None
        else:
            self.empty_label = empty_label
        self.cache_choices = cache_choices

        # Call Field instead of ChoiceField __init__() because we don't need
        # ChoiceField.__init__().
        Field.__init__(self, required, widget, label, initial, help_text,
                       *args, **kwargs)
        self.queryset = queryset
        self.choice_cache = None
        self.to_field_name = to_field_name

    def __deepcopy__(self, memo):
        result = super(ChoiceField, self).__deepcopy__(memo)
        # Need to force a new ModelChoiceIterator to be created, bug #11183
        result.queryset = result.queryset
        return result

    def _get_queryset(self):
        return self._queryset

    def _set_queryset(self, queryset):
        self._queryset = queryset
        self.widget.choices = self.choices

    queryset = property(_get_queryset, _set_queryset)

    # this method will be used to create object labels by the QuerySetIterator.
    # Override it to customize the label.
    def label_from_instance(self, obj):
        """
        This method is used to convert objects into strings; it's used to
        generate the labels for the choices presented by this object. Subclasses
        can override this method to customize the display of the choices.
        """
        return smart_unicode(obj)

    def _get_choices(self):
        # If self._choices is set, then somebody must have manually set
        # the property self.choices. In this case, just return self._choices.
        if hasattr(self, '_choices'):
            return self._choices

        # Otherwise, execute the QuerySet in self.queryset to determine the
        # choices dynamically. Return a fresh ModelChoiceIterator that has not been
        # consumed. Note that we're instantiating a new ModelChoiceIterator *each*
        # time _get_choices() is called (and, thus, each time self.choices is
        # accessed) so that we can ensure the QuerySet has not been consumed. This
        # construct might look complicated but it allows for lazy evaluation of
        # the queryset.
        return SchemaChoiceIterator(self)

    choices = property(_get_choices, ChoiceField._set_choices)

    def prepare_value(self, value):
        if hasattr(value, '_meta'):
            if self.to_field_name:
                return value.serializable_value(self.to_field_name)
            else:
                return value.pk
        return super(SchemaChoiceField, self).prepare_value(value)

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return None
        try:
            value = self.queryset.get(pk=value)
        except (ValueError):#, self.queryset.model.DoesNotExist): #TODO catch the propr does not exist error
            raise ValidationError(self.error_messages['invalid_choice'])
        return value

    def validate(self, value):
        return Field.validate(self, value)

class SchemaMultipleChoiceField(SchemaChoiceField):
    widget = SelectMultiple
    hidden_widget = MultipleHiddenInput
    default_error_messages = {
        'list': _(u'Enter a list of values.'),
        'invalid_choice': _(u'Select a valid choice. %s is not one of the'
                            u' available choices.'),
        'invalid_pk_value': _(u'"%s" is not a valid value for a primary key.')
    }

    def __init__(self, queryset, cache_choices=False, required=True,
                 widget=None, label=None, initial=None,
                 help_text=None, *args, **kwargs):
        super(SchemaMultipleChoiceField, self).__init__(queryset, None,
            cache_choices, required, widget, label, initial, help_text,
            *args, **kwargs)

    def clean(self, value):
        if self.required and not value:
            raise ValidationError(self.error_messages['required'])
        elif not self.required and not value:
            return []
        if not isinstance(value, (list, tuple, set)):
            raise ValidationError(self.error_messages['list'])
        
        #TODO validate the pk values
        '''
        qs = self.queryset.filter(**{'%s__in' % key: value})
        pks = set([force_unicode(getattr(o, key)) for o in qs])
        for val in value:
            if force_unicode(val) not in pks:
                raise ValidationError(self.error_messages['invalid_choice'] % val)
        '''
        # Since this overrides the inherited ModelChoiceField.clean
        # we run custom validators here
        self.run_validators(value)
        return value

    def prepare_value(self, value):
        if hasattr(value, '__iter__'):
            return [super(SchemaMultipleChoiceField, self).prepare_value(v) for v in value]
        return super(SchemaMultipleChoiceField, self).prepare_value(value)




from fields import BaseField as Field

from django.db.models.fields.files import FieldFile, FileDescriptor
from django.utils.encoding import force_unicode, smart_str
from django.utils.translation import ugettext_lazy
from django.core.files.storage import default_storage
from django.core.files import File
from django import forms

import datetime
import os

class FileField(Field):
    # The class to wrap instance attributes in. Accessing the file object off
    # the instance will always return an instance of attr_class.
    attr_class = FieldFile

    # The descriptor to use for accessing the attribute off of the class.
    descriptor_class = FileDescriptor

    description = ugettext_lazy("File path")
    
    form_field_class = forms.FileField

    def __init__(self, verbose_name=None, name=None, upload_to='', storage=None, **kwargs):
        self.storage = storage or default_storage
        self.upload_to = upload_to
        if callable(upload_to):
            self.generate_filename = upload_to

        super(FileField, self).__init__(verbose_name, name, **kwargs)

    def get_internal_type(self):
        return "FileField"

    #def contribute_to_class(self, cls, name):
    #    super(FileField, self).contribute_to_class(cls, name)
    #    setattr(cls, self.name, self.descriptor_class(self))

    def get_directory_name(self):
        return os.path.normpath(force_unicode(datetime.datetime.now().strftime(smart_str(self.upload_to))))

    def get_filename(self, filename):
        return os.path.normpath(self.storage.get_valid_name(os.path.basename(filename)))

    def generate_filename(self, filename):
        return os.path.join(self.get_directory_name(), self.get_filename(filename))

    def formfield(self, **kwargs):
        # If a file has been provided previously, then the form doesn't require
        # that a new file is provided this time.
        # The code to mark the form field as not required is used by
        # form_for_instance, but can probably be removed once form_for_instance
        # is gone. ModelForm uses a different method to check for an existing file.
        if 'initial' in kwargs:
            kwargs['required'] = False
        return super(FileField, self).formfield(**kwargs)
    
    def is_instance(self, val):
        return isinstance(val, File)
    
    def to_primitive(self, val):
        """
        Takes a python object and returns an object that is [json] serializable
        """
        if not val:
            return None
        if isinstance(val, basestring):
            return val
        if getattr(val, 'storage_path', False):
            return val.storage_path
        name = self.generate_filename(val.name)
        name = self.storage.save(name, val)
        assert name
        return name
    
    def to_python(self, val, parent=None):
        if not val:
            return None
        #self.attr_class(instance, field, name)
        ret = self.storage.open(val)
        ret.storage_path = val
        try:
            ret.url = self.storage.url(val)
        except Exception, error:
            print error
        return ret




from documentadmin import SchemaAdmin

from django import forms
from django.contrib.admin.util import flatten_fieldsets
from django.utils.functional import curry

from dockit.forms.formsets import BaseInlineFormSet, inlinedocumentformset_factory

class InlineSchemaAdmin(SchemaAdmin):
    schema = None
    formset = BaseInlineFormSet
    extra = 1
    max_num = None
    template = None
    verbose_name = None
    verbose_name_plural = None
    can_delete = True

    def __init__(self, model, admin_site, schema=None, documentadmin=None, dotpath=None, **kwargs):
        super(InlineSchemaAdmin, self).__init__(model, admin_site, schema or self.schema, documentadmin)
        self.model = self.schema
        self.opts = self.schema._meta
        self.dotpath = dotpath
        if self.verbose_name is None:
            self.verbose_name = self.opts.verbose_name
        if self.verbose_name_plural is None:
            #admin display hack, this or I copy the whole template
            field_name = self.dotpath
            if field_name.endswith('.*'):
                field_name = field_name[:-2]
            if '.' in field_name:
                field_name = field_name.rsplit('.',1)[-1]
            self.verbose_name_plural = field_name
        for key, value in kwargs.iteritems():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise TypeError("Unrecognized option: %s" % key)

    def _media(self):
        from django.conf import settings
        js = ['js/jquery.min.js', 'js/jquery.init.js', 'js/inlines.min.js']
        if self.prepopulated_fields:
            js.append('js/urlify.js')
            js.append('js/prepopulate.min.js')
        if self.filter_vertical or self.filter_horizontal:
            js.extend(['js/SelectBox.js' , 'js/SelectFilter2.js'])
        return forms.Media(js=['%s%s' % (settings.ADMIN_MEDIA_PREFIX, url) for url in js])
    media = property(_media)

    def get_formset(self, request, obj=None, **kwargs):
        """Returns a BaseInlineFormSet class for use in admin add/change views."""
        if self.declared_fieldsets:
            fields = flatten_fieldsets(self.declared_fieldsets)
        else:
            fields = None
        if self.exclude is None:
            exclude = []
        else:
            exclude = list(self.exclude)
        exclude.extend(kwargs.get("exclude", []))
        exclude.extend(self.get_readonly_fields(request, obj))
        # if exclude is an empty list we use None, since that's the actual
        # default
        exclude = exclude or None
        defaults = {
            "form": self.get_form_class(request, obj), #TODO this needs meta
            "formset": self.formset,
            "fields": fields,
            "exclude": exclude,
            "formfield_callback": curry(self.formfield_for_field, request=request),
            "extra": self.extra,
            "max_num": self.max_num,
            "can_delete": self.can_delete,
            "schema": self.schema,
        }
        defaults.update(kwargs)
        return inlinedocumentformset_factory(self.model, self.dotpath, **defaults)

    def get_fieldsets(self, request, obj=None):
        if self.declared_fieldsets:
            return self.declared_fieldsets
        form = self.get_formset(request).form
        fields = form.base_fields.keys() + list(self.get_readonly_fields(request, obj))
        return [(None, {'fields': fields})]

class StackedInline(InlineSchemaAdmin):
    template = 'admin/edit_inline/stacked.html'

class TabularInline(InlineSchemaAdmin):
    template = 'admin/edit_inline/tabular.html'




from django.db.models import Model, Q

from dockit.backends.indexer import BaseIndexer
from dockit.schema import fields#, Document

from models import DocumentStore, StringIndex
import models as indexes
from backend import ModelDocumentStorage, DocumentQuery

#TODO need a mechanism for back populating indexes, must be task based

class Indexer(object):
    def __init__(self, doc_class, index_creator, dotpath, name):
        self.doc_class = doc_class
        self.index_creator = index_creator
        self.dotpath = dotpath
        self.name = name
    
    def __call__(self, document):
        if self.dotpath in ('pk', '_pk'):
            return
        try:
            value = document.dot_notation(self.dotpath)
        except (KeyError, IndexError):
            return
        
        if value is None:
            return #TODO proper handling
        if isinstance(value, (list, set)):
            for val in value:
                self.index_creator(document.pk, self.name, val)
        else:
            self.index_creator(document.pk, self.name, value)

class ExactIndexer(BaseIndexer):
    INDEXES = [(fields.TextField, indexes.StringIndex),
           (fields.CharField, indexes.StringIndex),
           (fields.IntegerField, indexes.IntegerIndex),
           (fields.FloatField, indexes.FloatIndex),
           (fields.DecimalField, indexes.DecimalIndex),
           (fields.BooleanField, indexes.BooleanIndex),
           (fields.DateField, indexes.DateIndex),
           (fields.DateTimeField, indexes.DateTimeIndex),
           (fields.TimeField, indexes.TimeIndex),
           (Model, indexes.StringIndex),
           (fields.ReferenceField, indexes.StringIndex),
           (fields.ModelReferenceField, indexes.StringIndex),]
    
    def __init__(self, document, filter_operation):
        self.document = document
        self.filter_operation = filter_operation
        self.dotpath = self.filter_operation.dotpath()
        self.generate_index()
    
    def generate_index(self):
        collection = self.document._meta.collection
        field = self.document._meta.dot_notation_to_field(self.dotpath)
        
        subindex = self._lookup_index(field)
        if subindex is None and hasattr(field, 'subfield'):
            subindex = self._lookup_index(field.subfield)
        
        if subindex is None:
            subindex = StringIndex
            #raise TypeError("Could not identify an apropriate index for: %s" % field)
        
        self.subindex = subindex
        func = Indexer(self.document, subindex.objects.db_index, self.dotpath, self.filter_operation.key)
        filt = subindex.objects.filter_kwargs_for_operation
        unique_values = subindex.objects.unique_values
        clear = subindex.objects.clear_db_index
        
        self.index_functions = {'map':func, 'filter':filt, 'unique_values':unique_values, 'clear':clear}
    
    def _lookup_index(self, field):
        for key, val in self.INDEXES:
            if isinstance(field, key):
                return val
    
    def on_document_save(self, instance):
        self.index_functions['map'](instance)
    
    def on_document_delete(self, instance):
        self.index_functions['clear'](instance.pk)
        
    def filter(self):
        return Q(**self.index_functions['filter'](self.filter_operation))
    
    def values(self):
        return self.index_functions['unique_values'](self.filter_operation.key)

ModelDocumentStorage.register_indexer(ExactIndexer, 'exact', 'iexact', 'startswith', 'endswith', 'year', 'month', 'day')





class BaseDocumentStorage(object):
    _indexers = dict()
    
    @classmethod
    def register_indexer(cls, index_cls, *names):
        for name in names:
            cls._indexers[name] = index_cls
    
    @classmethod
    def get_indexer(cls, name):
        return cls._indexers[name]
    
    def _get_indexer_for_operation(self, document, op):
        indexer = self.get_indexer(op.operation)
        return indexer(document, op)
    
    def register_index(self, query_index):
        raise NotImplementedError
    
    def get_query(self, query_index):
        raise NotImplementedError
    
    def register_document(self, document):
        pass
        #for key, field in document._meta.fields.iteritems():
        #    if getattr(field, 'db_index', False):
        #        document.objects.enable_index("equals", key, {'field_name':key})
    
    def save(self, doc_class, collection, data):
        raise NotImplementedError
    
    def get(self, doc_class, collection, doc_id):
        '''
        Returns the primitive data for that doc_id
        '''
        raise NotImplementedError
    
    def delete(self, doc_class, collection, doc_id):
        raise NotImplementedError
    
    def get_id(self, data):
        return data.get(self.get_id_field_name())
    
    def get_id_field_name(self):
        raise NotImplementedError

class BaseDocumentQuerySet(object):
    def __len__(self):
        raise NotImplementedError
    
    def count(self):
        return self.__len__()
    
    def delete(self):
        raise NotImplementedError
    
    def all(self):
        from copy import copy
        return copy(self)
    
    def get(self, doc_id):
        raise NotImplementedError
    
    def __getitem__(self, val):
        raise NotImplementedError
    
    def __nonzero__(self):
        raise NotImplementedError
    
    def __and__(self, other):
        raise NotImplementedError
    
    def __or__(self, order):
        raise NotImplementedError




from copy import copy

from dockit.backends.queryindex import QueryIndex

class Manager(object):
    def contribute_to_class(self, cls, name):
        new = copy(self)
        new.schema = cls
        setattr(cls, name, new)
    
    @property
    def backend(self):
        return self.schema._meta.get_backend()
    
    @property
    def collection(self):
        return self.schema._meta.collection
    
    def filter(self, **kwargs):
        return self.all().filter(**kwargs)
    
    def index(self, *args):
        return self.all().index(*args)
    
    #def values(self):
    #    return self.index_manager.values
    
    def all(self):
        return QueryIndex(self.schema)
    
    def get(self, **kwargs):
        return self.all().get(**kwargs)

'''
register_indexer(backend, "equals", index_cls)

Book.objects.all().index('author_name__iexact').commit()
Book.objects.all().index('author_name__iexact').filter(author_name__iexact='Jane Doe')
#or
Book.objects.all().filter(author_name__iexact='Jane Doe')

Book.objects.all().index('author_name').values()
'''




from django.core.exceptions import ImproperlyConfigured
from django.views.generic import edit as editview

from detail import SingleObjectMixin, SingleObjectTemplateResponseMixin, BaseDetailView

from dockit.forms import DocumentForm

class DocumentFormMixin(editview.FormMixin, SingleObjectMixin):
    def get_form_class(self):
        """
        Returns the form class to use in this view
        """
        if self.form_class:
            return self.form_class
        else:
            if self.document is not None:
                # If a document has been explicitly provided, use it
                document = self.document
            elif hasattr(self, 'object') and self.object is not None:
                # If this view is operating on a single object, use
                # the class of that object
                document = self.object.__class__
            else:
                # Try to get a queryset and extract the document class
                # from that
                document = self.get_queryset().document
            #fields = fields_for_document(document)
            class CustomDocumentForm(DocumentForm):
                class Meta:
                    document = document
            #CustomDocumentForm.base_fields.update(fields)
            return CustomDocumentForm

    def get_form_kwargs(self):
        """
        Returns the keyword arguments for instanciating the form.
        """
        kwargs = super(DocumentFormMixin, self).get_form_kwargs()
        kwargs.update({'instance': self.object})
        return kwargs

    def get_success_url(self):
        if self.success_url:
            url = self.success_url % self.object.__dict__
        else:
            try:
                url = self.object.get_absolute_url()
            except AttributeError:
                raise ImproperlyConfigured(
                    "No URL to redirect to.  Either provide a url or define"
                    " a get_absolute_url method on the document.")
        return url

    def form_valid(self, form):
        self.object = form.save()
        return super(DocumentFormMixin, self).form_valid(form)

    def get_context_data(self, **kwargs):
        context = kwargs
        if getattr(self, 'object', None):
            context['object'] = self.object
            context_object_name = self.get_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        return context

class BaseCreateView(DocumentFormMixin, editview.ProcessFormView):
    """
    Base view for creating an new object instance.

    Using this base class requires subclassing to provide a response mixin.
    """
    def get(self, request, *args, **kwargs):
        self.object = None
        return super(BaseCreateView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = None
        return super(BaseCreateView, self).post(request, *args, **kwargs)


class CreateView(SingleObjectTemplateResponseMixin, BaseCreateView):
    """
    View for creating an new object instance,
    with a response rendered by template.
    """
    template_name_suffix = '_form'


class BaseUpdateView(DocumentFormMixin, editview.ProcessFormView):
    """
    Base view for updating an existing object.

    Using this base class requires subclassing to provide a response mixin.
    """
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseUpdateView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseUpdateView, self).post(request, *args, **kwargs)


class UpdateView(SingleObjectTemplateResponseMixin, BaseUpdateView):
    """
    View for updating an object,
    with a response rendered by template..
    """
    template_name_suffix = '_form'

DeletionMixin = editview.DeletionMixin

class BaseDeleteView(DeletionMixin, BaseDetailView):
    """
    Base view for deleting an object.

    Using this base class requires subclassing to provide a response mixin.
    """


class DeleteView(SingleObjectTemplateResponseMixin, BaseDeleteView):
    """
    View for deleting an object retrieved with `self.get_object()`,
    with a response rendered by template.
    """
    template_name_suffix = '_confirm_delete'




#!/usr/bin/env python

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

VERSION = '0.0.7'
LONG_DESC = """\
"""

setup(name='django-dockit',
      version=VERSION,
      description="",
      long_description=LONG_DESC,
      classifiers=[
          'Programming Language :: Python',
          'Operating System :: OS Independent',
          'Natural Language :: English',
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
      ],
      keywords='django',
      maintainer = 'Jason Kraus',
      maintainer_email = 'zbyte64@gmail.com',
      url='http://github.com/zbyte64/django-dockit',
      license='New BSD License',
      packages=find_packages(exclude=['test_environment']),
      #test_suite='tests.runtests.runtests',
      include_package_data = True,
      )



from django.utils import unittest

import dockit
from dockit import backends
from dockit.models import TemporaryDocument
from dockit.backends.mongo.backend import MongoDocumentStorage

class TestDocument(TemporaryDocument):
    charfield = dockit.CharField()
    listfield = dockit.ListField(dockit.CharField())

class MongoBackendTestCase(unittest.TestCase):
    def setUp(self):
        self._original_backend = backends.backend
        backends.backend = self._create_backend()
    
    def tearDown(self):
        backends.backend = self._original_backend
    
    def _create_backend(self):
        return MongoDocumentStorage(host='localhost', port=27017, db='testdb')
    
    def test_get_nonexistant_document_raises_error(self):
        try:
            TestDocument.objects.get(pk=None)
        except TestDocument.DoesNotExist:
            pass
        else:
            self.fail('Retrieved non-existant document')
        
        try:
            TestDocument.objects.get(pk='3f32ea6fd946e17d44000000')
        except TestDocument.DoesNotExist:
            pass
        else:
            self.fail('Retrieved non-existant document')
    
    def test_create_document(self):
        doc = TestDocument(charfield='test')
        doc.save()
        
        doc2 = TestDocument.objects.get(pk=doc.pk)
        self.assertEqual(doc, doc2)
        self.assertEqual(doc.charfield, doc2.charfield)
    
    def test_delete_document(self):
        doc = TestDocument(charfield='test')
        doc.save()
        pk = doc.pk
        doc.delete()
        try:
            doc2 = TestDocument.objects.get(pk=pk)
        except TestDocument.DoesNotExist:
            pass
        else:
            self.fail('Document was not deleted')
    
    def test_update_document(self):
        doc = TestDocument(charfield='test')
        doc.save()
        
        doc.charfield = 'test2'
        doc.listfield = ['a', 'b', 'c']
        doc.save()
        
        doc2 = TestDocument.objects.get(pk=doc.pk)
        self.assertEqual(doc.listfield, doc2.listfield)




from django.core.management.base import BaseCommand, CommandError
from django.core import serializers

from optparse import make_option

class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--format', default='json', dest='format',
            help='Specifies the output serialization format for fixtures.'),
        make_option('--indent', default=None, dest='indent', type='int',
            help='Specifies the indent level to use when pretty-printing output'),
    )
    help = ("Output the contents of the database as a fixture of the given "
            "format (using each model's default manager unless --all is "
            "specified).")
    #args = '[appname appname.ModelName ...]'

    def handle(self, *app_labels, **options):
        from dockit.schema.common import COLLECTIONS
        format = options.get('format','json')
        indent = options.get('indent',None)
        show_traceback = options.get('traceback', False)

        

        # Check that the serialization format exists; this is a shortcut to
        # avoid collating all the objects and _then_ failing.
        if format not in serializers.get_public_serializer_formats():
            raise CommandError("Unknown serialization format: %s" % format)

        try:
            serializers.get_serializer(format)
        except KeyError:
            raise CommandError("Unknown serialization format: %s" % format)

        # Now collate the objects to be serialized.
        objects = []
        excluded_documents = []
        for document in COLLECTIONS.itervalues():
            if document in excluded_documents:
                continue
            objects.extend(document.objects.all())

        try:
            return serializers.serialize(format, objects, indent=indent,
                        use_natural_keys=False)
        except Exception, e:
            if show_traceback:
                raise
            raise CommandError("Unable to serialize database: %s" % e)




from django.utils import simplejson
from django.core.serializers.json import DjangoJSONEncoder

from dockit.backends.base import BaseDocumentStorage
from dockit.backends.queryset import BaseDocumentQuery

from models import DocumentStore

class DocumentQuery(BaseDocumentQuery):
    def __init__(self, query_index, queryset):
        super(DocumentQuery, self).__init__(query_index)
        self.queryset = queryset
    
    def wrap(self, entry):
        data = simplejson.loads(entry.data)
        data['_pk'] = entry.pk
        return self.document.to_python(data)
    
    def delete(self):
        return self.queryset.delete()
    
    def get_from_filter_operations(self, filter_operations):
        queryset = self.queryset
        for op in filter_operations:
            indexer = self._get_indexer_for_operation(self.document, op)
            queryset = queryset.filter(indexer.filter())
        try:
            return self.wrap(queryset.get())
        except self.queryset.model.DoesNotExist:
            raise self.document.DoesNotExist
    
    def values(self):
        queryset = self.queryset
        for op in query_index.indexes:
            indexer = self._get_indexer_for_operation(self.document, op)
        raise NotImplementedError
    
    def __len__(self):
        return self.queryset.count()
    
    def __nonzero__(self):
        return bool(self.queryset)
    
    def __getitem__(self, val):
        if isinstance(val, slice):
            results = list()
            for entry in self.queryset[val]:
                results.append(self.wrap(entry))
            return results
        else:
            return self.wrap(self.queryset[val])

class ModelDocumentStorage(BaseDocumentStorage):
    name = "djangomodel"
    _indexers = dict() #TODO this should be automatic
    
    def __init__(self):
        self.indexes = dict()
        import indexers
    
    def get_id_field_name(self):
        return '_pk'
    
    def save(self, doc_class, collection, data):
        doc_id = self.get_id(data)
        encoded_data = simplejson.dumps(data, cls=DjangoJSONEncoder)
        document = DocumentStore(collection=collection, data=encoded_data)
        if doc_id is not None:
            document.pk = doc_id
        #CONSIDER this does not look before we save
        document.save()
        data[self.get_id_field_name()] = document.pk
        self._index_document(doc_class.to_python(data))
    
    def get(self, doc_class, collection, doc_id):
        try:
            document = DocumentStore.objects.get(collection=collection, pk=doc_id)
        except DocumentStore.DoesNotExist:
            raise doc_class.DoesNotExist
        data = simplejson.loads(document.data)
        data[self.get_id_field_name()] = document.pk
        return data
    
    def delete(self, doc_class, collection, doc_id):
        return DocumentStore.objects.filter(collection=collection, pk=doc_id).delete()
    
    def _index_document(self, instance):
        indexes = self.indexes.get(instance._meta.collection, set())
        for index in indexes:
            indexer = self._get_indexer_for_operation(type(instance), index)
            indexer.on_document_save(instance)
    
    def register_index(self, query_index):
        indexes = set(query_index.inclusions + query_index.exclusions + query_index.indexes)
        document = query_index.document
        self.indexes.setdefault(document._meta.collection, set())
        self.indexes[document._meta.collection].update(indexes)
    
    def get_query(self, query_index):
        document = query_index.document
        queryset = DocumentStore.objects.filter(collection=query_index.document._meta.collection)
        for op in query_index.inclusions:
            indexer = self._get_indexer_for_operation(document, op)
            queryset = queryset.filter(indexer.filter())
        for op in query_index.exclusions:
            indexer = self._get_indexer_for_operation(document, op)
            queryset = queryset.exclude(indexer.filter())
        return DocumentQuery(query_index, queryset)




from exceptions import DotPathNotFound

COLLECTIONS = dict()

def register_collection(cls):
    COLLECTIONS[cls._meta.collection] = cls

def get_base_document(key):
    return COLLECTIONS[key]

class UnSet(object):
    def __nonzero__(self):
        return False

class DotPathTraverser(object):
    def __init__(self, dotpath):
        self.dotpath = dotpath
        self.remaining_paths = dotpath.split('.')
        self.resolved_paths = list() #part, value, field
        self._finished = False
        self._called = False
    
    def resolve_for_schema(self, schema):
        from fields import SchemaField
        entry = {'value':None,
                 'field':SchemaField(schema=schema),
                 'part':None,}
        self.resolved_paths = [entry]
        self._resolve_loop()
    
    def resolve_for_instance(self, instance):
        from fields import SchemaField
        entry = {'value':instance,
                 'field':SchemaField(schema=type(instance)),
                 'part':None,}
        self.resolved_paths = [entry]
        self._resolve_loop()
    
    def _resolve_loop(self):
        while not self._finished:
            self._called = False
            self.resolve_next()
            if not self._called:
                assert False, str(self.resolved_paths)
            #need a better control routine, want one last resolve if end wasn't called
    
    def resolve_next(self):
        current = self.current
        if current['field'] is None:
            if hasattr(current['value'], 'traverse_dot_path'):
                current['value'].traverse_dot_path(self)
                #note the result may or may not have a field
            elif self.remaining_paths:
                raise DotPathNotFound('Arrived at a dead end', traverser=self)
            else:
                #nothing left to resolve
                self._finished = True
                self._called = True
        else:
            current['field'].traverse_dot_path(self)
    
    @property
    def next_part(self):
        return self.remaining_paths[0]
    
    @property
    def current(self):
        return self.resolved_paths[-1]
    
    @property
    def current_value(self):
        return self.current['value']
    
    def end(self, field=None, value=None):
        last_entry = self.resolved_paths[-1]
        if field:
            last_entry['field'] = field
        if value:
            last_entry['value'] = value
        assert len(self.remaining_paths) == 0
        self._finished = True
        self._called = True
    
    def next(self, field=None, value=None):
        part = self.remaining_paths.pop(0)
        if value and field is None:
            from fields import SchemaField
            from schema import Schema
            if isinstance(value, Schema):
                field = SchemaField(schema=type(value))
        entry = {'field':field,
                 'value':value,
                 'part':part,}
        self.resolved_paths.append(entry)
        self._called = True
    
    def set_value(self, value):
        parent_entry = self.resolved_paths[-2]
        part = self.current['part']
        if parent_entry['field']:
            parent_entry['field'].set_value(parent_entry['value'], part, value)
        elif parent_entry['value'] is not None:
            parent_entry['value'].set_value(part, value)
        else:
            assert False

class DotPathList(list):
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            new_value = None
            name = traverser.next_part
            try:
                new_value = self[int(name)]
            except ValueError:
                raise DotPathNotFound("Invalid index given, must be an integer")
            except IndexError:
                pass
            traverser.next(value=new_value)
        else:
            traverser.end(value=self)
    
    def set_value(self, attr, value):
        index = int(attr)
        if value is UnSet:
            self.pop(index)
        elif index == len(self):
            self.append(value)
        else:
            self[index] = value

class DotPathDict(dict):
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            new_value = None
            name = traverser.next_part
            try:
                new_value = self[name]
            except KeyError:
                pass
            traverser.next(value=new_value)
        else:
            traverser.end(value=self)
    
    def set_value(self, attr, value):
        if value is UnSet:
            del self[attr]
        else:
            self[attr] = value

class DotPathSet(set):
    def traverse_dot_path(self, traverser):
        if traverser.remaining_paths:
            raise DotPathNotFound("Cannot traverse past a set")
        else:
            traverser.end(value=self)
    
    def set_value(self, attr, value):
        if value is UnSet:
            self.remove(attr)
        else:
            self.add(value)

